#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
__doc__=r"""
:py:mod:`known/basic.py`
"""
#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
__all__ = [ 
    'Fake', 
    'Infinity', 
    'Kio', 
    'Remap',  
    'BaseConvert', 
    'IndexedDict', 
    'Verbose', 
    'Symbols', 
    'Table', 
    'Fuzz',
    ]
#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
import datetime, os
from math import floor, log, ceil
from collections import UserDict
from io import BytesIO
from zipfile import ZipFile
#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class Fake: # a fake object
    def __init__(self, **attributes):
        for name,value in attributes.items(): setattr(self, name, value)

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class Infinity: 
    # emulates infinity with comparision operators < <=, >, >=, ==
    #   +inf    =  Infinity(1)
    #   -inf    =  Infinity(-1)
    # emulates Empty and Universal sets with 'in' keyword
    #   Universal Set (everything) = Infinity(1)
    #   Null/Empty Set (nothing) = Infinity(-1)

    def __init__(self, sign=1) -> None: self.sign = (sign>=0) # has positive
    
    def __gt__(self, other): return self.sign       # +inf is greater than everything / -inf is greater than nothing
    def __ge__(self, other): return self.sign       

    def __lt__(self, other): return not self.sign   # -inf is less than everything    / +inf is less than nothing
    def __le__(self, other): return not self.sign

    def __eq__(self, other): return False           # inf is not equal to anything, not even itself

    def __contains__(self, x): return self.sign     # universal set contains everything (always true), empty set contains nothing (always false)

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class Kio:
    r""" provides input/out methods for loading saving python objects using json and pickle 
    Use methods:
        save_as_json(object, path)
        object = load_as_json(path)
        save_as_pickle(object, path)
        object = load_as_pickle(path)

    """

    import json, pickle
    IOFLAG = dict(
        json=   (json,      ''     ), 
        pickle= (pickle,    'b'    ),
        )

    @staticmethod
    def get_ioas(): return list(__class__.IOFLAG.keys())

    @staticmethod
    def save_buffer(o, ioas:str, seek0=False) -> None:
        assert ioas in __class__.IOFLAG, f'key error {ioas}'
        s_module, s_flag = __class__.IOFLAG[ioas]
        buffer = BytesIO()
        s_module.dump(o, buffer)
        if seek0: buffer.seek(0) # prepares for reading
        return buffer

    @staticmethod
    def load_buffer(buffer:BytesIO, ioas:str, seek0=True): 
        assert ioas in __class__.IOFLAG, f'key error {ioas}'
        s_module, s_flag = __class__.IOFLAG[ioas]
        if seek0: buffer.seek(0) # prepares for reading
        return s_module.load(buffer)

    @staticmethod
    def save_file(o, path:str, ioas:str, **kwargs):
        assert ioas in __class__.IOFLAG, f'key error {ioas}'
        s_module, s_flag = __class__.IOFLAG[ioas]
        with open(path, f'w{s_flag}') as f: s_module.dump(o, f, **kwargs)
        return path
    @staticmethod
    def load_file(path:str, ioas:str):
        assert ioas in __class__.IOFLAG, f'key error {ioas}'
        s_module, s_flag = __class__.IOFLAG[ioas]
        with open(path, f'r{s_flag}') as f: o = s_module.load(f)
        return o
    
    @staticmethod
    def save_as_json(o, path:str, **kwargs):    return __class__.save_file(o, path, 'json', **kwargs)
    @staticmethod
    def load_as_json(path:str):                     return __class__.load_file(path, 'json')

    @staticmethod
    def save_as_pickle(o, path:str, **kwargs):  return __class__.save_file(o, path, 'pickle', **kwargs)
    @staticmethod
    def load_as_pickle(path:str):                   return __class__.load_file(path, 'pickle')

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class Remap:
    r""" 
    Provides a mapping between ranges, works with scalars, ndarrays and tensors.

    :param Input_Range:     *FROM* range for ``forward`` call, *TO* range for ``backward`` call
    :param Output_Range:    *TO* range for ``forward`` call, *FROM* range for ``forward`` call
    """

    def __init__(self, Input_Range:tuple, Output_Range:tuple) -> None:
        r"""
        :param Input_Range:     `from` range for ``i2o`` call, `to` range for ``o2i`` call
        :param Output_Range:    `to` range for ``i2o`` call, `from` range for ``o2i`` call
        """
        self.set_input_range(Input_Range)
        self.set_output_range(Output_Range)

    def set_input_range(self, Range:tuple) -> None:
        r""" set the input range """
        self.input_low, self.input_high = Range
        self.input_delta = self.input_high - self.input_low

    def set_output_range(self, Range:tuple) -> None:
        r""" set the output range """
        self.output_low, self.output_high = Range
        self.output_delta = self.output_high - self.output_low

    def backward(self, X):
        r""" maps ``X`` from ``Output_Range`` to ``Input_Range`` """
        return ((X - self.output_low)*self.input_delta/self.output_delta) + self.input_low

    def forward(self, X):
        r""" maps ``X`` from ``Input_Range`` to ``Output_Range`` """
        return ((X - self.input_low)*self.output_delta/self.input_delta) + self.output_low

    def __call__(self, X, backward=False):
        return self.backward(X) if backward else self.forward(X)
    
    def swap_range(self):
        Input_Range, Output_Range = (self.output_low, self.output_high), (self.input_low, self.input_high)
        self.set_input_range(Input_Range)
        self.set_output_range(Output_Range)

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class BaseConvert:
    
    r""" Number System Conversion 
    
    A number is abstract concept that has many representations using sets of symbols

    A base-n number system uses a set of n digits to represent any number
    This is called the representation of the number

    Given one representation, we only need to convert to another

    """

    @staticmethod
    def zeros(n): return [0 for _ in range(n)]

    @staticmethod
    def convert(digits, base_from, base_to, reversed=True):
        r""" convers from one base to another 
        
        :param digits:      iterable of digits in base ```base_from```. NOTE: digits are Natural Numbers starting at 0. base 'b' will have digits between [0, b-1]
        :param base_from:   int - the base to convert from
        :param base_to:     int - the base to convert to
        :param reversed:    bool - if True, digits are assumed in reverse (human readable left to right)
                            e.g. if reversed is True then binary digits iterable [1,0,0] will represent [4] in decimal otherwise it will represent [1] in decimal
        """

        digits_from =  [int(abs(d)) for d in digits] # convert to int data-type
        if reversed: digits_from = digits_from[::-1]
        ndigits_from = len(digits_from)
        mult_from = [base_from**i for i in range(ndigits_from)]
        repr_from = sum([ui*vi for ui,vi in zip(digits_from,mult_from, strict=True)]) #dot(digits_from , mult_from)

        #ndc = base_from**ndigits_from
        ndigits_to = ceil(log(repr_from,base_to))
        digits_to =  __class__.zeros(ndigits_to) 
        n = int(repr_from)
        for d in range(ndigits_to):
            digits_to[d] = n%base_to
            n=n//base_to

        if reversed: digits_to = digits_to[::-1]
        return tuple(digits_to)


    @staticmethod
    def ndigits(num:int, base:int): return ceil(log(num,base))

    @staticmethod
    def int2base(num:int, base:int, digs:int) -> list:
        r""" 
        Convert base-10 integer to a base-n list of fixed no. of digits 

        :param num:     base-10 number to be represented
        :param base:    base-n number system
        :param digs:    no of digits in the output

        :returns:       represented number as a list of ordinals in base-n number system

        .. seealso::
            :func:`~known.basic.base2int`
        """
        
        ndigits = digs if digs else ceil(log(num,base)) 
        digits =  __class__.zeros(ndigits)
        n = num
        for d in range(ndigits):
            digits[d] = n%base
            n=n//base
        return digits

    @staticmethod
    def base2int(num, base:int) -> int:
        """ 
        Convert an iterbale of digits in base-n system to base-10 integer

        :param num:     iterable of base-n digits
        :param base:    base-n number system

        :returns:       represented number as a integer in base-10 number system

        .. seealso::
            :func:`~known.basic.int2base`
        """
        res = 0
        for i,n in enumerate(num): res+=(base**i)*n
        return int(res)


    SYM_BIN = { f'{i}':i for i in range(2) }
    SYM_OCT = { f'{i}':i for i in range(8) }
    SYM_DEC = { f'{i}':i for i in range(10) }
    SYM_HEX = {**SYM_DEC , **{ s:(i+10) for i,s in enumerate(('A', 'B', 'C', 'D', 'E', 'F'))}}
    
    @staticmethod
    def n_syms(n): return { f'{i}':i for i in range(n) }

    @staticmethod
    def to_base_10(syms:dict, num:str):
        b = len(syms)
        l = [ syms[n] for n in num[::-1] ]
        return __class__.base2int(l, b)

    @staticmethod
    def from_base_10(syms:dict, num:int, joiner='', ndigs=None):
        base = len(syms)
        #print(f'----{num=} {type(num)}, {base=}, {type(base)}')
        if not ndigs: ndigs = (1 + (0 if num==0 else floor(log(num, base))))  # __class__.ndigs(num, base)
        ss = tuple(syms.keys())
        S = [ ss[i]  for i in __class__.int2base(num, base, ndigs) ]
        return joiner.join(S[::-1])

    @staticmethod
    def int2hex(num:int, joiner=''): return __class__.from_base_10(__class__.SYM_HEX, num, joiner)

    @staticmethod
    def int2bases(integer, bases): # Generalized Case: (int=i, bases=[b1,...,bn]) -> digits=[d1,...,dn]
        r""" converts integer to n digits, each from n bases """
        digits = []
        q = int(integer)
        for b in bases:
            q, r = divmod(q, b)
            digits.append(r)
        return digits

    @staticmethod
    def bases2int(bases, digits): # Generalized Case: (bases=[b1,...,bn], digits=[d1,...,dn])
        r""" converts n digits, each from n bases to a integer """
        m,n = 1,0
        for b,d in zip(bases, digits):
            n+=(d*m)
            m*=b
        return n

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class IndexedDict(UserDict):
    r""" Implements an Indexed dict where values can be addressed using both index(int) and keys(str) """

    def __init__(self, **members) -> None:
        self.names = []
        super().__init__(*[], **members)
    
    def keys(self): return enumerate(self.names, 0) # for i,k in self.keys()

    def items(self): return enumerate(self.data.items(), 0) # for i,(k,v) in self.items()

    def __len__(self): return len(self.data)

    def __getitem__(self, name): 
        if isinstance(name, int): name = self.names[name]
        if name in self.data: 
            return self.data[name]
        else:
            raise KeyError(name)

    def __setitem__(self, name, item): 
        if isinstance(name, int): name = self.names[name]
        if name not in self.data: self.names.append(name)
        self.data[name] = item

    def __delitem__(self, name): 
        index = None
        if isinstance(name, int):  
            index = name
            name = self.names[name]
        if name in self.data: 
            del self.names[self.names.index(name) if index is None else index]
            del self.data[name]

    def __iter__(self): return iter(self.names)

    def __contains__(self, name): return name in self.data

    # Now, add the methods in dicts but not in MutableMapping

    def __repr__(self) -> str:
        return f'{__class__} :: {len(self)} Members'
    
    def __str__(self) -> str:
        items = ''
        for i,k in enumerate(self):
            items += f'[{i}] \t {k} : {self[i]}\n'
        return f'{__class__} :: {len(self)} Members\n{items}'
    
    def __copy__(self):
        inst = self.__class__.__new__(self.__class__)
        inst.__dict__.update(self.__dict__)
        # Create a copy and avoid triggering descriptors
        inst.__dict__["data"] = self.__dict__["data"].copy()
        inst.__dict__["names"] = self.__dict__["names"].copy()
        return inst

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class Table:
    r""" a simple table data-structure implement using python dict with disk IO """

    @staticmethod
    def _Create(*columns):
        data = {None:[f'{col}' for col in columns]} # this is to make sure that col names are always on top
        return data

    @staticmethod
    def Create(columns:tuple, primary_key:str, cell_delimiter=',', record_delimiter='\n'):
        # should be called on a new object after init\
        table = __class__()
        table.data = __class__._Create(*columns)
        table.pk = primary_key
        table.pkat = table.data[None].index(table.pk)
        table.cell_delimiter, table.record_delimiter = cell_delimiter, record_delimiter
        return table


    @staticmethod
    def _Import(path, key_at, cell_delimiter, record_delimiter): 
        with open(path, 'r', encoding='utf-8') as f: 
            s = f.read()
            lines = s.split(record_delimiter)
            cols = lines[0].split(cell_delimiter) #<--- only if None:cols was added as a first entry (using Create method)
            data = {None:cols}
            if isinstance(key_at, str): key_at = cols.index(key_at)
            assert key_at>=0,f'Invlaid key {key_at}'
            for line in lines[1:]:
                if line:
                    cells = line.split(cell_delimiter)
                    data[f'{cells[key_at]}'] = cells
        return data
    
    @staticmethod
    def Import(path, key_at, cell_delimiter=',', record_delimiter='\n'): 
        table = __class__()
        table.data = __class__._Import(path, key_at, cell_delimiter, record_delimiter)
        if isinstance(key_at, str): key_at = table[None].index(key_at)
        table.pk = table.data[None][key_at]
        table.pkat = key_at
        table.cell_delimiter, table.record_delimiter = cell_delimiter, record_delimiter
        return table


    @staticmethod
    def _Export(data, path, cell_delimiter, record_delimiter): 
        with open(path, 'w', encoding='utf-8') as f: 
            for v in data.values(): f.write(cell_delimiter.join(v)+record_delimiter)

    @staticmethod
    def Export(table, path): 
        __class__._Export(table.data, path, table.cell_delimiter, table.record_delimiter)

    @property
    def cols(self): return self[None]

    @property
    def keys(self): return set([k for k in self.data.keys() if k is not None])

    # get row as dict
    def __call__(self, key): return {k:v for k,v in zip(self[None], self[key])}

    # get row as it is (list)
    def __getitem__(self, key): return self.data[key]

    # set row based on if its a dict or a list (note: key is irrelavant here)
    def __setitem__(self, key, row):
        assert key is not None, f'Cannot set `None` key'
        assert isinstance(key, str), f'Expected string type keys but got {type(key)}'
        if isinstance(row, dict):
            # dict does not require to match exact length
            assert self.pk not in row , f'Cannot change primary key' 
            indexof = {k:self[None].index(k) for k in row}
            if not key in self.data: self[key] = None 
            self.data[key][self.pkat] = key
            for k,v in row.items(): self.data[key][indexof[k]] = str(v)
            

        elif isinstance(row, (list, tuple)):
            # must exactly match no of cols, 
            # use ... to keep values unchanged 
            # value are initialized to None
            assert len(row) == len(self[None]), f'Rows are expected to have length {len(self[None])} but got {len(row)}'
            assert row[self.pkat] is ... , f'Cannot change primary key'
            if not key in self.data: self[key] = None 
            self.data[key][self.pkat] = key
            for i,v in enumerate(row):
                if not (v is ...) : self.data[key][i] = str(v)
        elif row is None: 
            # (re)initializes the row
            self.data[key] = ['' for _ in self[None]]
            self.data[key][self.pkat] = key
        else: raise AssertionError(f'Excepting a dict/list/tuple/None but got {type(row)}')

    # del row based on key
    def __delitem__(self, key):
        if key is not None: del self.data[key]

    def __contains__(self, key): return key in self.data

    # quick export > file
    def __gt__(self, other):__class__._Export(self.data, f'{other}', self.cell_delimiter, self.record_delimiter)

    # quick import < file
    def __lt__(self, other): self.data = __class__._Import(f'{other}', self.pkat, self.cell_delimiter, self.record_delimiter)

    # total number of rows
    def __len__(self): return len(self.data)-1

    def __str__(self):
        return ...

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class Verbose:
    r""" Contains shorthand helper functions for printing outputs and representing objects as strings.
            Methods ending with '_' in name return Strings instead of printing them
    """
    
    #-----------------------------------------------------------------------
    """ SECTION: HRS - Human Readable String  for sizes"""
    #-----------------------------------------------------------------------
    HRS_MAPPER = dict(BB=2**0, KB=2**10, MB=2**20, GB=2**30, TB=2**40) # 2 chars for keys

    @staticmethod
    def hrs2bytes(hrsize:str): return round(float(hrsize[:-2])*__class__.HRS_MAPPER.get(hrsize[-2:].upper(), 0))
    @staticmethod
    def bytes2hrs(size:int, unit:str, roundoff=2): return f"{round(size/(__class__.HRS_MAPPER[unit]),roundoff)}{unit}"

    @staticmethod
    def bytes2bb(size:int, roundoff=2): return __class__.bytes2hrs(size, 'BB', roundoff)
    @staticmethod
    def bytes2kb(size:int, roundoff=2): return __class__.bytes2hrs(size, 'KB', roundoff)
    @staticmethod
    def bytes2mb(size:int, roundoff=2): return __class__.bytes2hrs(size, 'MB', roundoff)
    @staticmethod
    def bytes2gb(size:int, roundoff=2): return __class__.bytes2hrs(size, 'GB', roundoff)
    @staticmethod
    def bytes2tb(size:int, roundoff=2): return __class__.bytes2hrs(size, 'TB', roundoff)

    @staticmethod
    def bytes2auto(size:int, roundoff=2):
        if      size<__class__.HRS_MAPPER["KB"]: return __class__.bytes2bb(size, roundoff)
        elif    size<__class__.HRS_MAPPER["MB"]: return __class__.bytes2kb(size, roundoff)
        elif    size<__class__.HRS_MAPPER["GB"]: return __class__.bytes2mb(size, roundoff)
        elif    size<__class__.HRS_MAPPER["TB"]: return __class__.bytes2gb(size, roundoff)
        else                                   : return __class__.bytes2tb(size, roundoff)
    #-----------------------------------------------------------------------


    #-----------------------------------------------------------------------
    """ SECTION: StrX - human readable string representation of objects """
    #-----------------------------------------------------------------------

    DEFAULT_DATE_FORMAT = ["%Y","%m","%d","%H","%M","%S","%f"] #  Default date format for :func:`~known.basic.Verbose.strU` 
    DASHED_LINE = "=-=-=-=-==-=-=-=-="

    @staticmethod
    def strN(s:str, n:int) -> str:  
        r""" Repeates a string n-times """
        return ''.join([s for _ in range(n)])

    @staticmethod
    def _recP_(a, level, index, pindex, tabchar='\t', show_dim=False):
        # helper function for recP - do not use directly
        if index<0: index=''
        dimstr = ('* ' if level<1 else f'*{level-1} ') if show_dim else ''
        pindex = f'{pindex}{index}'
        if len(a.shape)==0:
            print(f'{__class__.strN(tabchar, level)}[ {dimstr}@{pindex}\t {a} ]') 
        else:
            print(f'{__class__.strN(tabchar, level)}[ {dimstr}@{pindex} #{a.shape[0]}')
            for i,s in enumerate(a):
                __class__._recP_(s, level+1, i, pindex, tabchar, show_dim)
            print(f'{__class__.strN(tabchar, level)}]')

    @staticmethod
    def recP(arr, show_dim:bool=False) -> None: 
        r"""
        Recursive Print - print an iterable recursively with added indentation.

        :param arr:         any iterable with ``shape`` property.
        :param show_dim:    if `True`, prints the dimension at the start of each item
        """
        __class__._recP_(arr, 0, -1, '', '\t', show_dim)
    
    @staticmethod
    def strA_(arr, start:str="", sep:str="|", end:str="") -> str:
        r"""
        String Array - returns a string representation of an iterable for printing.
        
        :param arr:     input iterable
        :param start:   string prefix
        :param sep:     item seperator
        :param end:     string postfix
        """
        res=start
        for a in arr: res += (str(a) + sep)
        return res + end

    @staticmethod
    def strA(arr, start:str="", sep:str="|", end:str="") -> None: print(__class__.strA_(arr, start, sep, end))
    
    @staticmethod
    def strD_(arr, sep:str="\n", cep:str=":\n", caption:str="") -> str:
        r"""
        String Dict - returns a string representation of a dict object for printing.
        
        :param arr:     input dict
        :param sep:     item seperator
        :param cep:     key-value seperator
        :param caption: heading at the top
        """
        res=f"=-=-=-=-==-=-=-=-={sep}DICT #[{len(arr)}] : {caption}{sep}{__class__.DASHED_LINE}{sep}"
        for k,v in arr.items(): res+=str(k) + cep + str(v) + sep
        return f"{res}{__class__.DASHED_LINE}{sep}"

    @staticmethod
    def strD(arr, sep:str="\n", cep:str=":\n", caption:str="") -> None: print(__class__.strD_(arr, sep, cep, caption))

    @staticmethod
    def strU(form, start:str='', sep:str='', end:str='') -> str:
        r""" 
        String UID - returns a formated string of current timestamp.

        :param form: the format of timestamp, If `None`, uses the default :data:`~known.basic.Verbose.DEFAULT_DATE_FORMAT`.
            Can be selected from a sub-set of ``["%Y","%m","%d","%H","%M","%S","%f"]``.
            
        :param start: UID prefix
        :param sep: UID seperator
        :param end: UID postfix

        """
        if not form: form = __class__.DEFAULT_DATE_FORMAT
        return start + datetime.datetime.strftime(datetime.datetime.now(), sep.join(form)) + end

    @staticmethod
    def now(year:bool=True, month:bool=True, day:bool=True, 
            hour:bool=True, minute:bool=True, second:bool=True, mirco:bool=True, 
            start:str='', sep:str='', end:str='') -> str:
        r""" Unique Identifier - useful in generating unique identifiers based on current timestamp. 
        Helpful in generating unique filenames based on timestamps. 
        
        .. seealso::
            :func:`~known.basic.Verbose.strU`
        """
        form = []
        if year:    form.append("%Y")
        if month:   form.append("%m")
        if day:     form.append("%d")
        if hour:    form.append("%H")
        if minute:  form.append("%M")
        if second:  form.append("%S")
        if mirco:   form.append("%f")
        assert (form), 'format should not be empty!'
        return (start + datetime.datetime.strftime(datetime.datetime.now(), sep.join(form)) + end)


    #-----------------------------------------------------------------------
    """ SECTION: show/info - human readable information about pbjects """
    #-----------------------------------------------------------------------

    DOCSTR_FORM = lambda x: f'\t!docstr:\n! - - - - - - - - - - - - - - - - -\n{x}\n- - - - - - - - - - - - - - - - - !'

    @staticmethod
    def show_(x, cep:str='\t\t:', sep="\n", sw:str='__', ew:str='__') -> None:
        res = ""
        for d in dir(x):
            if not (d.startswith(sw) or d.endswith(ew)):
                v = ""
                try:
                    v = getattr(x, d)
                except:
                    v='?'
                res+=f'({d} {cep} {v}{sep}'
        return res

    @staticmethod
    def show(x, cep:str='\t\t:', sep="\n", sw:str='__', ew:str='__') -> None:
        r"""
        Show Object - describes members of an object using the ``dir`` call.

        :param x:       the object to be described
        :param cep:     the name-value seperator
        :param sw:      argument for ``startswith`` to check in member name
        :param ew:      argument for ``endswith`` to check in member name

        .. note:: ``string.startswith`` and ``string.endswith`` checks are performed on each member of the object 
            and only matching member are displayed. This is usually done to prevent showing dunder members.
        
        .. seealso::
            :func:`~known.basic.Verbose.dir`
        """
        print(__class__.show_(x, cep=cep, sw=sw, ew=ew))

    @staticmethod
    def dir(x, doc=False, filter:str='', sew=('__','__')):
        """ Calls ```dir``` on given argument and lists the name and types of non-dunder members.

        :param filter: csv string of types to filter out like `type,function,module`, keep blank for no filter
        :param doc: shows docstring ```__doc``` 
            If ```doc``` is True, show all member's ```__doc__```.
            If ```doc``` is False, does not show any ```__doc__```. 
            If ```doc``` is a string, show ```__doc__``` of specific types only given by csv string.

        :param sew: 2-Tuple (start:str, end:str) - excludes member names that start and end with specific chars, 
            used to exclude dunder methods by default

        .. seealso::
            :func:`~known.basic.Verbose.show`
        """
        #if self_doc: print( f'{type(x)}\n{x.__doc__}\n' )
        if sew: sw, ew = f'{sew[0]}', f'{sew[1]}'
        doc_is_specified = (isinstance(doc, str) and bool(doc))
        if doc_is_specified: doc_match =[ t for t in doc.replace(' ','').split(',') if t ]
        if filter: filter_match =[ t for t in filter.replace(' ','').split(',') if t ]
        counter=1
        for k in dir(x):
            if sew:
                if (k.startswith(sw) and k.endswith(ew)): continue
            m = getattr(x,k)
            n = str(type(m)).split("'")[1]
            if filter:
                if not (n in filter_match):  continue
            s = f'[{counter}] {k} :: {n}'#.encode('utf-16')

            if doc:
                if doc_is_specified:
                    if n in doc_match: 
                        d = __class__.DOCSTR_FORM(m.__doc__)
                    else:
                        d=''
                else:
                    d = __class__.DOCSTR_FORM(m.__doc__)
            else:
                d = ''
            counter+=1
            print(f'{s}{d}')

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class Symbols:
    CORRECT =       '✓'
    INCORRECT =     '✗'
    ALPHA =         'α'
    BETA =          'β'
    GAMMA =         'γ'
    DELTA =         'δ'
    EPSILON =       'ε'
    ZETA =          'ζ'
    ETA =           'η'
    THETA =         'θ'
    KAPPA =         'κ'
    LAMBDA =        'λ'
    MU =            'μ' 
    XI =            'ξ'
    PI =            'π'
    ROH =           'ρ'
    SIGMA =         'σ'
    PHI =           'φ'
    PSI =           'Ψ'
    TAU =           'τ'
    OMEGA =         'Ω'
    TRI =           'Δ'
    DOT=            '●'
    SUN=            '⚙'
    ARROW1=         '↦'
    ARROW2=         '⇒'
    ARROW3=         '↪'

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class Fuzz:
    r""" file system related """

    @staticmethod
    def ExistInfo(path):
        """ returns 4-tuple (exists,isdir,isfile,islink) """
        return os.path.exists(path), os.path.isdir(path), os.path.isfile(path), os.path.islink(path)
    
    @staticmethod
    def Scan(path, exclude_hidden=False, include_size=False, include_extra=False):
        #if not os.path.exists(path): return []
        r""" Scans a directory using os.scandir call 
             returns a list of 6 or 9 tuple (name, path, isdir, isfile, islink, size, parent, fname, ext) """
        #print(f'\t{path=}\n\t{exclude_hidden=}\n\t{include_size=}\n\t{include_extra=}\n')
        if exclude_hidden:  
            if include_size: 
                if include_extra:   return [(x.name, os.path.abspath(x.path), x.is_dir(), x.is_file(), x.is_symlink(), x.stat().st_size, os.path.dirname(os.path.abspath(x.path)), *__class__.SplitName(x.name)) for x in os.scandir(path) if not x.name.startswith(".")]
                else:               return [(x.name, os.path.abspath(x.path), x.is_dir(), x.is_file(), x.is_symlink(), x.stat().st_size) for x in os.scandir(path) if not x.name.startswith(".")]

            else:            
                if include_extra:   return [(x.name, os.path.abspath(x.path), x.is_dir(), x.is_file(), x.is_symlink(), -1              , os.path.dirname(os.path.abspath(x.path)), *__class__.SplitName(x.name)) for x in os.scandir(path) if not x.name.startswith(".")]
                else:               return [(x.name, os.path.abspath(x.path), x.is_dir(), x.is_file(), x.is_symlink(), -1              ) for x in os.scandir(path) if not x.name.startswith(".")]
        else:
            if include_size: 
                if include_extra:   return [(x.name, os.path.abspath(x.path), x.is_dir(), x.is_file(), x.is_symlink(), x.stat().st_size, os.path.dirname(os.path.abspath(x.path)), *__class__.SplitName(x.name)) for x in os.scandir(path)]
                else:               return [(x.name, os.path.abspath(x.path), x.is_dir(), x.is_file(), x.is_symlink(), x.stat().st_size) for x in os.scandir(path)]
            else:
                if include_extra:   return [(x.name, os.path.abspath(x.path), x.is_dir(), x.is_file(), x.is_symlink(), -1              ,os.path.dirname(os.path.abspath(x.path)), *__class__.SplitName(x.name)) for x in os.scandir(path)]
                else:               return [(x.name, os.path.abspath(x.path), x.is_dir(), x.is_file(), x.is_symlink(), -1              ) for x in os.scandir(path)]

    @staticmethod
    def ReScan(path, exclude_hidden=False, include_size=False, include_extra=False):
        r""" Recursively Scans a directory using os.scandir """
        res = []
        pending = [path]
        while pending:
            try:
                ls = __class__.Scan(pending.pop(0), exclude_hidden=exclude_hidden, include_size=include_size, include_extra=include_extra)
                for l in ls: # name, path, isdir, isfile, islink, size, parent, (fname, ext)
                    res.append(l)
                    if l[2]: (pending.append(l[1]) if not l[4] else None)  #if l[2]: (pending.append(l[1]) if not l[4] else (pending.append(l[1]) if expand_links else None))  
            except: pass
        return res

    @staticmethod
    def SplitFileName(path:str): return __class__.SplitName(os.path.basename(path))
        
    @staticmethod
    def SplitName(f:str):
        r"""splits a file-name into name.ext 
        Note: make sure to pass os.path.basename() to this function
        Retutns: 2-tuple (name, ext)
            `name` is always a string
            `ext` can be None or a string (None means that there is no "." in file name)
        """
        i = f.rfind('.')
        return (f, None) if i<0 else (f[0:i], f[i+1:])

    @staticmethod
    def RenameFile(path, new_name, keep_ext=False):
        """ rename a file with or without changing its extension """
        dirname, filename = os.path.dirname(path), os.path.basename(path)
        _, ext = __class__.SplitName(f'{filename}')
        if keep_ext and (ext is not None):  name_ = f'{new_name}.{ext}'
        else:                               name_ = f'{new_name}'
        return os.path.join(dirname, name_)
    
    @staticmethod
    def RenameExt(path, new_ext):
        """ change a file extension without renaming it """
        dirname, filename = os.path.dirname(path), os.path.basename(path)
        name, _ = __class__.SplitName(f'{filename}')
        return os.path.join(dirname, f'{name}.{new_ext}')

    @staticmethod
    def Zip(nested:bool, zpath:str, *paths):
        if not (paths and zpath): return None
        if not zpath.lower().endswith('.zip'): zpath = f'{zpath}.zip' 
        zipn = zpath[:-4]
        paths = list(paths)
        tozip = {}
        while paths:
            path = paths.pop(0)
            
            if isinstance(path, (list, tuple)): item_path, arc_path = path
            else: item_path, arc_path = path, path

            if os.path.islink(item_path): continue
            isfile, isdir = os.path.isfile(item_path), os.path.isdir(item_path)
            if not (isfile or isdir): continue
            arc_path = os.path.basename(arc_path)
            
            if isfile:
                assert arc_path not in tozip, f'Duplicate names detected {arc_path} from {path}'
                tozip[arc_path] = os.path.abspath(item_path)
            else:
                for l in __class__.ReScan(item_path): 
                    arc = os.path.join(arc_path, os.path.relpath( l[1], item_path))
                    assert arc not in tozip, f'Duplicate names detected {arc} from {path}'
                    tozip[arc] = l[1]

        with ZipFile(zpath, 'w') as zip_object:
            if nested: zip_object.mkdir(zipn)
            for arc_path, file_path in tozip.items():
                if nested: arc_path = os.path.join(zipn, arc_path)
                zip_object.write(filename=file_path, arcname=arc_path)
        return tozip

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=


__doc__="""Nothing useful here - Contains temporary/old code snippets, classes, functions etc"""







#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

# class Fancy(Mailer):
#     r""" A fancy email notifier that uses operators to send mail in one line

#     example:
#     (__class__() \
#         / 'from@gmail.com' \
#         // 'password' \
#         * 'subject' \
#         @ 'to@gmail.com' \
#         % 'cc@gmail.com' \
#         - 'Email Msg Body' \
#         + 'attachment.txt')()

#     # NOTE: to read sender credentials from env-variables start the script using custom env as:
#     # env "USERNAME=username@gmail.com" "PASSWORD=???? ???? ???? ????" python run.py

#     # NOTE: the "joiner" arg specifies the char used to join list items in body

#     # NOTE: Operator preceedence 
#     # * @ / // %
#     # + -
#     # << >>
#     # & ^ | 

#     """
#     def __init__(self, joiner='\n'): 
#         self.joiner = joiner
#         self.subject, self.to, self.cc, self.username, self.password  = '', '', '', '', ''
#         self.content, self.attached = [], set([])
#         self._status = False
        
#     def write(self, *lines): 
#         self.content.extend(lines)
#         return self

#     def attach(self, *files):
#         self.attached = self.attached.union(set(files))
#         return self
    
#     # Level 1 --------------------------------------------------------------

#     def __truediv__(self, username):     # username/FROM     nf / 'from_mail@gmail.com'
#         self.username = username  
#         return self
    
#     def __floordiv__(self, password):    # password         nf // 'pass word'
#         self.password = password 
#         return self
    
#     def __mul__(self, subject):  # subject      nf * "subject"
#         self.subject = subject   
#         return self

#     def __matmul__(self, to_csv):  # TO         nf @ 'to_mail@gmail.com,...'
#         self.to = to_csv  
#         return self
    
#     def __mod__(self, cc_csv):      # CC       nf % 'cc_mail@gmail.com,...'
#         self.cc = cc_csv
#         return self

#     # Level 2 --------------------------------------------------------------

#     def __sub__(self, content):   # body        nf - "content"
#         self.write(content)        
#         return self
    
#     def __add__(self, file):     # attachement      nf + "a.txt"
#         self.attach(file)        
#         return self

#     # Level 3 (SPECIAL CASES ONLY) -----------------------------------------

#     def __invert__(self): return self.Compose(
#             From=self.username,
#             Subject=self.subject,
#             To= self.to,
#             Cc= self.cc,
#             Body=self.joiner.join(self.content),
#             Attached=tuple(self.attached),
#         ) # composing ~nf

#     def __and__(self, username): # set username     nf & "username"
#         self.username = username  
#         return self

#     def __xor__(self, password): # set password     nf ^ "password"
#         self.password = password 
#         return self

#     def __or__(self, other):    # send mail         nf | 1
#         if other: self._status = self()
#         else: self._status = False
#         return self

#     def __bool__(self): return self._status

#     # Level 4 --------------------------------------------------------------

#     def __call__(self, msg=None): return self.Send(
#         msg = (msg if msg else ~self),
#         username=( self.username if self.username else os.environ.get('USERNAME', '') ),
#         password=( self.password if self.password else os.environ.get('PASSWORD', '') ),
#         )

#     #--------------------------------------------------------------









# class AutoFetcher:
#     r""" Fetches emails from a folder in gmail and transfers them to a Queue, 
#         then processes this Queue using user-defined Callback 
#         A callback can be defined in the child class
#         def Callback(self, From, To, Cc, Bcc, Subject, Body, Attachements, Alias, **kwargs): return ...
#         """

#     def __init__(self):
#         self.mailboxes =                        {} # mailboxes to monitor
#         self.Q =                                [] # processing queue

#     def Create(self, alias, username, password):
#         self.mailboxes[alias] = Mailbox().Setup(username=username, password=password, alias=alias)
#         return self

#     def Fetch(self, alias, folder, save="", delete=False):
#         # get all messages from given folder (and flag as well)
#         mailbox = self.mailboxes[alias]
#         try:    _= mailbox.Login()
#         except: return False, f'Cannot open connection'
        
#         if mailbox.imap.state != 'AUTH': return False, f'Cannot authorize connection'

#         reason, rstatus = mailbox.OpenFolder(folder)
#         if not rstatus: return False, f'Cannot open {folder}: {reason}'

#         criteria = ['ALL'] if delete else ['UNFLAGGED']

#         mstatus, mcount = mailbox.GetMessageList(criteria=criteria)
#         if mstatus != 'OK': return False, f'Cannot get message list from {folder}: {reason}'
        
#         for i in range(mcount): self.Q.append(mailbox.GetMessage(i, save=save, seen=False, flag=not delete, delete=delete))        
#         mailbox.CloseFolder()
#         mailbox.Logout()
#         return True, f'Added {mcount} tasks'

#     def Work(self, popat=0):
#         while self.Q: status = self.Callback(**self.Q.pop(popat))
            
#     def Callback(self, **kwargs): return ...
#     #def Callback(self, From, To, Cc, Bcc, Subject, Body, Attachements, Alias, **kwargs): return ...

# #=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=



#-----------------------------------------------------------------------------------------
from sys import exit
if __name__!='__main__': exit(f'[!] can not import {__name__}.{__file__}')
#-----------------------------------------------------------------------------------------
#%% Arg-parse
import os, argparse, datetime, logging
# python -m known.nbserver --help
parser = argparse.ArgumentParser()


parser.add_argument('--base',           type=str, default='',           help="path to base dir, defaults to current directory")
parser.add_argument('--template',       type=str, default='lab',        help="classic/[lab]/reveal")
parser.add_argument('--home',           type=str, default='',           help="home page if not specified, creates a new notebook with the same name as the base dir")
parser.add_argument('--noscript',       type=int, default=0,            help="[default=0] if =1, remove any embedded <script> tags (may break latex/mathjax)")
parser.add_argument('--nofiles',        type=int, default=0,            help="[default=0] if =1, prevents downloading files - only notebooks can be viewed and downloaded ")
parser.add_argument('--log',            type=str, default='',           help="log file name - keep empty for no logging")
parser.add_argument('--host',           type=str, default='0.0.0.0',    help="ip-Interface, keep 0.0.0.0 for all interfaces")
parser.add_argument('--port',           type=str, default='8080',       help="port")
parser.add_argument('--threads',        type=int, default=10,           help="no of threads used by waitress server")
parser.add_argument('--max_con',        type=int, default=500,          help="maximum number of connections allowed to the server")
parser.add_argument('--max_size',       type=str, default='1024MB',     help="maximum size of http body")
# Notebook decorations
parser.add_argument('--rtext',          type=str, default='♻️',         help="text for refresh link")      
parser.add_argument('--dtext',          type=str, default='📥️',         help="text for download link")       
parser.add_argument('--ttext',          type=str, default='🔝',         help="text for go-top link")
parser.add_argument('--htext',          type=str, default='🏠',         help="text for home link")
parser.add_argument('--ltext',          type=str, default='🔒',         help="text for logout link")
parser.add_argument('--halign',         type=str, default='left',       help="header alignment")
parser.add_argument('--header',         type=int, default=0,            help="shows text in the header")

# extra login
parser.add_argument('--login',        type=str, default='',           help="path to login-dir, each user has a file with password inside, keep blank to disbale authentication")
parser.add_argument('--secret',       type=str, default='',           help="the secret for flask, required only when login is enbaled")
parser.add_argument('--case',         type=int, default=0,            help="uid case sensetivity (-1 = to-lower) (0 = no-change) (1 = to-upper)")
parser.add_argument('--welcome',      type=str, default='Welcome',    help="welcome msg shown on the login page")
parser.add_argument('--https',        type=int, default=0,            help="if True, Tells waitress that its behind an nginx proxy - https://flask.palletsprojects.com/en/stable/deploying/nginx/")
parsed = parser.parse_args()

if parsed.login:
    LOGIN_XL_PATH = os.path.abspath(parsed.login)
    try:os.makedirs(LOGIN_XL_PATH, exist_ok=True)
    except:pass
    if not os.path.isdir(LOGIN_XL_PATH): exit(f'Login-dir not found at {LOGIN_XL_PATH}')
else: LOGIN_XL_PATH = None

#%% Logging
LOGFILE = f'{parsed.log}'
if LOGFILE: 
    try:
        logging.basicConfig(filename=LOGFILE, filemode='a', level=logging.INFO, format='%(asctime)s - %(message)s')
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.INFO)
        formatter = logging.Formatter('%(asctime)s - %(message)s')
        console_handler.setFormatter(formatter)
        logger = logging.getLogger()
        logger.addHandler(console_handler)
    except: exit(f'[!] Logging could not be setup at {LOGFILE}')
    def sprint(msg): logging.info(msg) 
else:
    def sprint(msg): print(msg) 

#%% imports
PROXY_FIX=bool(parsed.https)
import nbconvert, os
from bs4 import BeautifulSoup
from flask import Flask, request, abort, redirect, url_for, send_file, session, render_template_string
if PROXY_FIX: from werkzeug.middleware.proxy_fix import ProxyFix
from waitress import serve
#from flask import Flask, render_template, request, redirect, url_for, session, abort, send_file
#%% Common
str2bytes_sizes = dict(BB=2**0, KB=2**10, MB=2**20, GB=2**30, TB=2**40)
def str2bytes(size): return int(float(size[:-2])*str2bytes_sizes.get(size[-2:].upper(), 0))

style_actions = """
.btn_header {
    background-color: #FFFFFF; 
    margin: 0px 0px 0px 6px;
    padding: 12px 6px 12px 6px;
    border-style: solid;
    border-width: thin;
    border-color: #000000;
    color: #000000;
    font-weight: bold;
    font-size: medium;
    border-radius: 5px;
}

.btn_actions {
    background-color: #FFFFFF; 
    padding: 2px 2px 2px 2px;
    margin: 5px 5px 5px 5px;
    border-style: solid;
    border-color: silver;
    border-width: thin;
    color: #000000;
    font-weight: bold;
    font-size: medium;
    border-radius: 2px;
}


"""
import re

def rematch(instr, pattern):  return \
    (len(instr) >= 0) and \
    (len(instr) <= 50) and \
    (re.match(pattern, instr))

def VALIDATE_PASS(instr):     return rematch(instr, r'^[a-zA-Z0-9~!@#$%^&*()_+{}<>?`\-=\[\].]+$')
def VALIDATE_UID(instr):      return rematch(instr, r'^[a-zA-Z0-9._@]+$') and instr[0]!="."


def get_passwd(dir, user):
    try:
        f = open(os.path.join(dir, user), 'r')
        passwd = f.read()
        f.close()
    except: passwd = None
    return passwd

def set_passwd(dir, user, passwd, create=False):
    try:
        p = os.path.join(dir, user)
        if not os.path.isfile(p): assert create
        f = open(p, 'w')
        f.write(passwd)
        f.close()
        return True
    except: return False

LOGIN_NEED_TEXT =       '👤' 
LOGIN_FAIL_TEXT =       '❌'     
LOGIN_NEW_TEXT =        '🔥'
LOGIN_CREATE_TEXT =     '🔑' 

login = """
<html>
    <head>
        <meta charset="UTF-8">
        <title> {{ config.title }} </title>
        <link rel="icon" href="/favicon.ico">
    </head>
    <body>
    <!-- ---------------------------------------------------------->
    </br>
    <!-- ---------------------------------------------------------->

    <div align="center">
        <br>
        <form action="{{ url_for('route_login') }}" method="post">
            <br>
            <div style="font-size: x-large;">{{ warn }}</div>
            <br>
            <div style="font-size: x-large; font-family: monospace">{{ msg }}</div>
            <br>
            <input id="uid" name="uid" type="text"  style="font-size: medium; font-family: monospace; text-align: center;" placeholder=" user-id " />
            <br>
            <br>
            <input id="passwd" name="passwd"   style="font-size: medium; font-family: monospace; text-align: center;" type="password" placeholder=" password " />
            <br>
            <br>
            <input type="submit" style="font-size: large;" value=""" + f'"{parsed.ltext}"' +"""> 
            <br>
            <br>
        </form>
    </div>

    <!-- ---------------------------------------------------------->
    
    <br>
    </div>
    <!-- ---------------------------------------------------------->
    </body>
</html>
"""

def nb2html(source_notebook, html_title=None, 
            template_name='lab', no_script=False, favicon=True, 
            auth=False, 
            llink='logout', 
            hlink='home', 
            header=0, 
            tlink='top', 
            clink='clear', curl='#',
            rlink='refresh', rurl='#', 
            dlink='download', durl='#', 
            align='left'):
    # ==============================================================
    if html_title is None:
        html_title = os.path.basename(source_notebook)
        if html_title.lower().endswith(".ipynb"): html_title = html_title[:-6]
    page, _ = nbconvert.HTMLExporter(template_name=template_name) \
            .from_file(source_notebook, dict(metadata=dict(name = f'{html_title}')),) 
    soup = BeautifulSoup(page, 'html.parser')
    # ==============================================================
    
    if no_script:
        for script in soup.find_all('script'): script.decompose()  # Find all script tags and remove them
    
    if favicon:
        link_tag = soup.new_tag('link')
        link_tag['rel'] = 'icon'
        link_tag['href'] = 'favicon.ico'
        soup.head.insert(0, link_tag)

    #if tlink or hlink or dlink or header: 
    style_tag = soup.new_tag('style')
    style_tag['type'] = 'text/css'
    style_tag.string = style_actions
    soup.head.insert(0, style_tag)


    #if hlink or dlink or header:
    ndiv = soup.new_tag('div')
    ndiv['align'] = f'{align}'
    html_string = ""
    if auth: html_string += f'<a class="btn_actions" href="/logout">{llink}</a>'
    if hlink: html_string += f'<a class="btn_actions" href="/">{hlink}</a>' 
    if clink: html_string += f'<a class="btn_actions" href="{curl}">{clink}</a>' 
    if dlink: html_string += f'<a class="btn_actions" href="{durl}">{dlink}</a>' 
    if rlink: html_string += f'<a class="btn_actions" href="{rurl}">{rlink}</a>' 
    if header: html_string += f'<span class="btn_header">{html_title} @ ./{os.path.relpath(source_notebook, app.config["base"])}</span>'
    html_string += f'<br>'
    nstr = BeautifulSoup(html_string, 'html.parser')
    ndiv.append(nstr) 
    soup.body.insert(0, ndiv)

    if tlink:
        ndiv = soup.new_tag('div')
        ndiv['align'] = f'{align}'
        html_string = f'<hr><a class="btn_actions" href="#">{tlink}</a><br>'
        nstr = BeautifulSoup(html_string, 'html.parser')
        ndiv.append(nstr) 
        soup.body.append(ndiv)

    # ==============================================================
    # final_page = soup.prettify()
    # ==============================================================
    return soup.prettify()

def new_notebook(heading="", nbformat=4, nbformat_minor=2):
    return '{"cells": [{"cell_type": "markdown","metadata": {},"source": [ "# '+str(heading)+'" ] } ], "metadata": { }, "nbformat": '+str(nbformat)+', "nbformat_minor": '+str(nbformat_minor)+'}'

#%% App Setup 


BASE = os.path.abspath(parsed.base)
try: os.makedirs(BASE, exist_ok=True)
except: 
    sprint(f'No directory found at {BASE}, using current directory...')
    BASE = os.path.abspath(os.getcwd())

sprint(f'⇒ Serving from directory {BASE}')
EXT = ".ipynb"
PH = f'{parsed.home}'
if not PH: PH=os.path.basename(BASE)
HOME = f'{PH}{EXT}'
HOME_PATH = os.path.join(BASE, HOME)
if not os.path.isfile(HOME_PATH):
    try: 
        with open(HOME_PATH, 'w') as f: f.write(new_notebook(os.path.basename(BASE)))
    except: exit(f'The home page at {HOME_PATH} was not found and could not be created.')
if not os.path.isfile(HOME_PATH): exit(f'Home page "{HOME}" not found at {HOME_PATH}.')

AUTH_ON = (LOGIN_XL_PATH is not None)
SECRET_KEY = parsed.secret
if AUTH_ON and not SECRET_KEY: exit(f'Secret Key was not provided!')

app = Flask(__name__, static_folder=BASE, template_folder=BASE)
if PROXY_FIX: app.wsgi_app = ProxyFix(app.wsgi_app, x_for=1, x_proto=1, x_host=1, x_prefix=1)
if AUTH_ON: app.secret_key =  SECRET_KEY
app.config['auth'] = AUTH_ON
app.config['base'] = BASE
app.config['template'] = parsed.template
app.config['dtext'] = parsed.dtext
app.config['rtext'] = parsed.rtext
app.config['ttext'] = parsed.ttext
app.config['htext'] = parsed.htext
app.config['ltext'] = parsed.ltext
app.config['halign'] = parsed.halign
app.config['header'] = int(parsed.header)
app.config['home'] = HOME
app.config['title'] = os.path.basename(BASE)
app.config['ext'] = EXT
app.config['query_refresh'] = "!"
app.config['query_download'] = "?"
app.config['query_clear'] = "~"
app.config['no_script'] = bool(parsed.noscript)
app.config['no_files'] = bool(parsed.nofiles)

loaded_pages = dict()

#%% Routes Section
#from known import Verbose as vb
@app.route('/', methods =['GET'], defaults={'query': ''})
@app.route('/<path:query>')
def route_home(query):
    if app.config['auth']:
        if not session.get('has_login', False): return redirect(url_for('route_login'))
    #sprint("\n-------------------------------------------------------------")
    #sprint("[NEWREQUEST]")
    #sprint("-------------------------------------------------------------\n")
    #sprint(vb.show_(request))
    #sprint("\n-------------------------------------------------------------\n")
    refresh = app.config['query_refresh'] in request.args
    download = app.config['query_download'] in request.args
    clear = app.config['query_clear'] in request.args
    base, ext, home = app.config['base'], app.config['ext'], app.config['home']
    tosend = False
    
    if ('.' in os.path.basename(query)):    tosend = (not query.lower().endswith(ext))
    else:                                   query += ext #---> auto add extension
    if ext==query: 			                query=home

    showdlink = not((query==home) or (query==ext))
    #sprint (f'[{"🔸" if showdlink else "🔹"}]{request.remote_addr}\t[{request.method}] {request.url}')
    from_name = session["uid"] if app.config['auth'] else "👤"
    sprint (f'🔸 {from_name}\t({request.remote_addr})\t[{request.method}] {request.url}') # \n{request.headers}
    #sprint("__________________________________________________________________\n")
    requested = os.path.join(base, query) # Joining the base and the requested path
    if not ((os.path.isfile(requested)) and (not os.path.relpath(requested, base).startswith(base))): return abort(404)
    else:
        if tosend: 
            if requested.endswith("favicon.ico"): return send_file(requested, as_attachment=False) 
            else: return abort(403) if app.config['no_files'] else send_file(requested, as_attachment=False) 
        else:
            global loaded_pages
            if clear and not showdlink: # clear before loading
                loaded_pages.clear()
                return redirect(url_for('route_home'))
            if (requested not in loaded_pages) or refresh: loaded_pages[requested] = nb2html(
                    requested, 
                    html_title=app.config['title'] if not showdlink else None, 
                    template_name=app.config['template'], 
                    no_script=app.config['no_script'], 
                    favicon=True, 
                    auth = app.config['auth'],
                    llink=app.config['ltext'], 
                    tlink=app.config['ttext'], 
                    rlink=app.config['rtext'] if showdlink else None, 
                    dlink=app.config['dtext'] if showdlink else None, 
                    hlink = app.config['htext'] if showdlink else None,
                    clink = app.config['rtext'] if not showdlink else None,
                    header = app.config['header'] if showdlink else None,
                    durl=f"{request.path}?{app.config['query_download']}", 
                    rurl=f"{request.path}?{app.config['query_refresh']}", 
                    curl=f"{request.path}?{app.config['query_clear']}", 
                    align=app.config['halign'])
                #with open('??.html','w') as f: f.write(loaded_pages[requested]) # save a copy to disk?
            if refresh: return redirect(url_for('route_home', query=query))
            else:
                if download: return send_file(requested, as_attachment=True) #<--- downloading ipynbs
                else: return  loaded_pages[requested]


#%% TEMP ROUTES

# ------------------------------------------------------------------------------------------
# login
 
# ------------------------------------------------------------------------------------------
@app.route('/login', methods =['GET', 'POST'])
def route_login():
    if not app.config['auth']: 
        session['has_login'], session['uid'] = True, 'no-auth'
        return redirect(url_for('route_home'))
    if request.method == 'POST' and 'uid' in request.form and 'passwd' in request.form:
        in_uid = f"{request.form['uid']}"
        in_passwd = f"{request.form['passwd']}"

        uid = in_uid if not parsed.case else (in_uid.upper() if parsed.case>0 else in_uid.lower())
        valid_query = VALIDATE_UID(uid)
        if not valid_query : passwd = None
        else: passwd = get_passwd(LOGIN_XL_PATH, uid)
        if passwd is not None: 
            if not passwd: # fist login
                #warn = LOGIN_NEW_TEXT
                #msg = f'[{in_uid}] has not set their password'
                           
                if in_passwd: # new password provided
                    if VALIDATE_PASS(in_passwd): # new password is valid
                        set_passwd(LOGIN_XL_PATH, uid, in_passwd, create=False)
                        warn = LOGIN_CREATE_TEXT
                        msg = f'[{in_uid}] New password was created successfully'
                        sprint(f'● {in_uid} just joined via {request.remote_addr}')
           
                    else: # new password is invalid valid 
                        warn = LOGIN_NEW_TEXT
                        msg=f'[{in_uid}] New password is invalid - can use any of the alphabets (A-Z, a-z), numbers (0-9), underscore (_), dot (.) and at-symbol (@) only'
                        
                                               
                else: #new password not provided                
                    warn = LOGIN_NEW_TEXT
                    msg = f'[{in_uid}] New password required - can use any of the alphabets (A-Z, a-z), numbers (0-9), underscore (_), dot (.) and at-symbol (@) only'
                                                     
            else: # re login
                if in_passwd: # password provided 
                    if in_passwd==passwd:
                        session['has_login'], session['uid'] = True, uid
                        sprint(f'● {session["uid"]} has logged in via {request.remote_addr}') 
                        return redirect(url_for('route_home'))
                    else:  
                        warn = LOGIN_FAIL_TEXT
                        msg = f'[{in_uid}] Password mismatch'                  
                else: # password not provided
                    warn = LOGIN_FAIL_TEXT
                    msg = f'[{in_uid}] Password not provided'
        else:
            warn = LOGIN_FAIL_TEXT
            msg = f'[{in_uid}] Not a valid user' 

    else:
        
        if session.get('has_login', False):  return redirect(url_for('route_home'))
        msg = parsed.welcome
        warn = LOGIN_NEED_TEXT 
        
    return render_template_string(login, msg = msg,  warn = warn)
# ------------------------------------------------------------------------------------------

# logout
# ------------------------------------------------------------------------------------------
@app.route('/logout')
def route_logout():
    r""" logout a user and redirect to login page """
    if not app.config['auth']: return redirect(url_for('route_home'))
    if not session.get('has_login', False):  return redirect(url_for('route_login'))
    if not session.get('uid', False): return redirect(url_for('route_login'))
    if session['has_login']:  sprint(f'● {session["uid"]} has logged out via {request.remote_addr}') 
    else: sprint(f'✗ {session["uid"]} was removed due to invalid uid ({session["uid"]}) via {request.remote_addr}') 
    session.clear()
    return redirect(url_for('route_login'))
# ------------------------------------------------------------------------------------------

#%% Server Section
def endpoints(athost):
    if athost=='0.0.0.0':
        ips=set()
        try:
            import socket
            for info in socket.getaddrinfo(socket.gethostname(), None):
                if (info[0].name == socket.AddressFamily.AF_INET.name): ips.add(info[4][0])
        except: pass
        ips=list(ips)
        ips.extend(['127.0.0.1', 'localhost'])
        return ips
    else: return [f'{athost}']

for endpoint in endpoints(parsed.host): sprint(f'◉ http://{endpoint}:{parsed.port}')
start_time = datetime.datetime.now()
sprint('◉ start server @ [{}]'.format(start_time))
serve(app,
    host = parsed.host,          
    port = parsed.port,          
    url_scheme = 'http',     
    threads = parsed.threads,    
    connection_limit = parsed.max_con,
    max_request_body_size = str2bytes(parsed.max_size),
    _quiet=True,
)
end_time = datetime.datetime.now()
sprint('◉ stop server @ [{}]'.format(end_time))
sprint('◉ server up-time was [{}]'.format(end_time - start_time))

#%%

# author: Nelson.S



__doc__=r"""email tools"""

__all__ = [ 
    'Mailer', 
    'Mailbox', 
    ]

import os
import smtplib, mimetypes, imaplib
from email import message_from_bytes
from email.message import EmailMessage
from email.header import decode_header

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class Mailer:
    r"""
    Use gmail account to read and send mails.
    Use `SendMail` static method to send mail
    Use `ReadMail` static method to read mail
    Both these methods require username and password
        > username must be a gmail address from which mail will be sent - recivers can be in any domain
        > password can be either one of:
            > google (account) password
            > app password

    Its recomended to use app password, to generate one visit 
        > https://myaccount.google.com/apppasswords
    
    App passwords requires enabling 2-step verification first, to enable it visit
        > https://myaccount.google.com/signinoptions/twosv

    
    -> on the reciver side - mark incoming mails as'not spam' at least once

    """
    @staticmethod
    def Mimes(files, default_ctype='application/octet-stream'):
        r""" gets mimetype info all files in a list """
        if isinstance(files, str): files=[f'{files}']
        res = []
        for path in files:
            if not os.path.isfile(path): continue
            ctype, encoding = mimetypes.guess_type(path)
            if ctype is None or encoding is not None: ctype = default_ctype
            maintype, subtype = ctype.split('/', 1)
            res.append( (path, maintype, subtype) )
        return res
    
    @staticmethod
    def Compose(From:str, Subject:str, To:str, Cc:str, Body:str, Attached:list=None, html:bool=False):
        msg = EmailMessage()

        # from, subject and recivers
        msg['From'] = f'{From}' 
        msg['Subject'] = f'{Subject}'
        msg['To'] = To
        if Cc: msg['Cc'] = Cc
        msg.set_content(Body)
        if html: msg.add_alternative(Body, subtype='html')
        # attachement
        if Attached is None: Attached = []
        assert isinstance(Attached, (list, tuple)), f'Expecting a list or tuple of attached files but got {type(Attached)}'       
        for file_name,main_type,sub_type in __class__.Mimes(Attached):
            with open (file_name, 'rb') as f: file_data = f.read()
            msg.add_attachment(file_data, 
                                maintype=main_type, 
                                subtype=sub_type, 
                                filename=os.path.basename(file_name))

        return msg

    @staticmethod
    def Send(msg:EmailMessage, username:str, password:str, url='smtp.gmail.com', port='587', tls=True):
        r""" send a msg using url:port with provided credentials, calls `starttls` is `tls` is True """
        if not (username and password): return False
        try:
            with smtplib.SMTP(f'{url}', int(port)) as smtp: 
                if tls: smtp.starttls()
                # assert username == msg['From'] #<--- should be?
                smtp.login(username, password) 
                smtp.ehlo()
                smtp.send_message(msg)
        except: return False
        return True

    @staticmethod
    def SendMail(
        username, 
        password, 
        Subject, 
        To, 
        Cc, 
        Body, 
        Attached:list=None,
        html:bool=False,
        url='smtp.gmail.com', port='587', tls=True,
    ): return __class__.Send(__class__.Compose(username, Subject, To, Cc, Body, Attached=Attached, html=html), 
            username, password, url=url, port=port, tls=tls)

    @staticmethod
    def ReadFolders(
        username, 
        password, 
    ):
        imap = imaplib.IMAP4_SSL("imap.gmail.com")
        imap.login(username, password)
        status, folders = imap.list()
        folders = [folder.decode().split('"')[-2] for folder in folders]
        imap.logout()
        return status, tuple(folders)

    @staticmethod
    def ReadMail(
        username, 
        password, 
        folder='INBOX',
        criteria='ALL',
        fetch='BODY[]',
        save='',
        delete=False,
        force=False,
    ):
        r"""

        #------------------------------------------
        Argument `ReadMail.folder`
        #------------------------------------------
        status, folders = imap.list()
        #------------------------------------------


        #------------------------------------------
        Argument `ReadMail.criteria`
        #------------------------------------------
        "ALL"	                        All messages
        "UNSEEN"	                    Unread messages
        "SEEN"	                        Read messages
        "FROM \"someone@example.com\""	Messages from a specific sender
        "TO \"someone@example.com\""	Sent to specific recipient
        "SUBJECT \"word\""	            Subject contains "word"
        "BODY \"word\""	                Body contains "word"
        "SINCE 01-Jan-2024"	            Messages after this date
        "BEFORE 01-Jun-2025"	        Messages before this date
        #------------------------------------------

        
        #------------------------------------------
        Argument `ReadMail.fetch`
        #------------------------------------------
        "BODY[]"	    Same as "RFC822" — full message
        "BODY[HEADER]"	Only the email headers (no body)
        "BODY[TEXT]"	Only the plain body, no headers
        "BODYSTRUCTURE"	Describes MIME structure (useful for parsing)
        "FLAGS"	        Read/unread flags (e.g. \Seen)
        "ENVELOPE"	    Summary info like subject, from, to, etc.
        #------------------------------------------


        """
        save = os.path.abspath(save) if save else None
        if delete and not save: 
            print(f'Warning: deleting without saving...')
            if not force: raise ZeroDivisionError(f'... set force=True to delete messages without saving')
                
        imap = imaplib.IMAP4_SSL("imap.gmail.com")
        imap.login(username, password)
        # To list other folders use: status, folders = imap.list()
        imap.select(folder)  # folder="INBOX"
        bstatus, bmessages = imap.search(None, criteria)
        messages = bmessages[0].split()
        bmsg=[]

        while messages:
            latest_msg = messages.pop()
            
            # Fetch the email
            status, msg_data = imap.fetch(latest_msg, fetch) # fetch="(RFC822)"
            for response_part in msg_data:
                if isinstance(response_part, tuple):
                    msg = message_from_bytes(response_part[1])
                
                    # Decode subject
                    subject, encoding = decode_header(msg["Subject"])[0]
                    if isinstance(subject, bytes):
                        subject = subject.decode(encoding or "utf-8")
                    
                    attachments=[]
                    # Process email parts
                    for part in msg.walk():
                        content_type = part.get_content_type()
                        content_dispo = str(part.get("Content-Disposition"))

                        # Get the plain text body
                        if content_type == "text/plain" and "attachment" not in content_dispo:
                            body = part.get_payload(decode=True).decode(errors="ignore")

                        # Handle attachments
                        
                        if "attachment" in content_dispo:
                            filename = part.get_filename()
                            
                            if filename and save:
                                decoded_filename, enc = decode_header(filename)[0]
                                if isinstance(decoded_filename, bytes):
                                    filename = decoded_filename.decode(enc or "utf-8")
                                
                                filepath = os.path.join(save, filename)
                                with open(filepath, "wb") as f:
                                    f.write(part.get_payload(decode=True))
                                print(f"Attachment saved: {filepath}")
                            else:
                                filepath=None
                            
                            attachments.append((filename, filepath))
                    # 
                    bmsg.append( {
                    'From' : msg.get("From"),
                    'To' : msg.get("To"),
                    'Cc' : msg.get("Cc"),
                    'Bcc' : msg.get("Bcc"),
                    'Subject': subject,
                    'Body': body,
                    'Attachements': attachments,
                    })
            #
            if delete: imap.store(latest_msg, '+FLAGS', '\\Deleted')

        #
        if delete: imap.expunge()       
        imap.close()       
        imap.logout()

        return bstatus, bmsg

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class Mailbox:
    r""" Email Reader based on imap.gmail """

    def __str__(self):return f'[{self.alias}] ({self.username})'
    def __repr__(self):return str(self)

    def Setup(self,
        username, 
        password,
        alias="",
        server='imap.gmail.com',
        port='993',
    ):
        self.alias=f'{alias}'
        self.username=f'{username}'
        self.password=f'{password}'
        self.server=server
        self.port=port
        self.Reset()
        return self
        
    def Reset(self):
        self.folder = None  # after self.OpenFolder
        self.messages = None # after self.GetMessageList 
        self.imap = None 

    def Login(self):
        self.imap = imaplib.IMAP4_SSL(self.server, int(self.port) )
        self.imap.login(self.username, self.password)
        return self.imap.state, self.imap.welcome

    def ListFolders(self):
        if self.imap.state == 'AUTH':
            status, folders = self.imap.list()
            return status, [folder.decode().split('"')[-2] for folder in folders]
        else:  return f'[State is {self.imap.state}] cannot list folders', None

    def OpenFolder(self, folder='INBOX'):
        self.imap.select(folder)
        if self.imap.state=='SELECTED': 
            self.folder=folder
            return f'[State is {self.imap.state}] selected folder ({folder})', True
        else: return f'[State is {self.imap.state}] cannot select folder ({folder})', False

    def GetMessageList(self, criteria=['ALL',],):
        if self.imap.state=='SELECTED':
            bstatus, bmessages = self.imap.search(None, *criteria)
            self.messages = bmessages[0].split()
            return bstatus, len(self.messages)
        else: return f'[State is {self.imap.state}] cannot get messages', -1

    def GetMessageInfo(self):
        if self.imap.state=='SELECTED':
            res=[]
            if self.messages is not None:
                for message_id in self.messages:
                    estatus, edata = self.imap.fetch(message_id, 'ENVELOPE')
                    fstatus, fdata = self.imap.fetch(message_id, 'FLAGS')
                    edata=edata[0].decode()
                    fdata=fdata[0].decode()
                    res.append((edata, fdata))
                return True, res
            else: return False, f'[State is {self.imap.state}] message list not avaibale, call GetMessageList first'
        else: return False, f'[State is {self.imap.state}] cannot open messages'

    def GetMessage(self,
        index,
        save='',
        seen=False,
        flag=False,
        delete=False,
    ):
        message_id = self.messages[index]
        status, msg_data = self.imap.fetch(message_id, 'BODY[]') # fetch="(RFC822)"
        # ----------------------------------------------------
        msg = message_from_bytes(msg_data[0][1])
        subject, encoding = decode_header(msg["Subject"])[0]
        if isinstance(subject, bytes):subject = subject.decode(encoding or "utf-8")
        attachments=[]
        body=""
        save = os.path.abspath(save) if save else None
        #if delete and not save: print(f'Warning: deleting without saving attachements ...')
        for part in msg.walk():
            content_type = part.get_content_type()
            content_dispo = str(part.get("Content-Disposition"))
            if content_type == "text/plain" and "attachment" not in content_dispo:
                body += part.get_payload(decode=True).decode(errors="ignore")            
            if "attachment" in content_dispo:
                filename = part.get_filename()
                if filename and save:
                    decoded_filename, enc = decode_header(filename)[0]
                    if isinstance(decoded_filename, bytes): filename = decoded_filename.decode(enc or "utf-8")
                    filepath = os.path.join(save, filename)
                    with open(filepath, "wb") as f: f.write(part.get_payload(decode=True))
                else: filepath=None
                attachments.append((filename, filepath))
            # ----------------------------------------------------
        res = {
        'From' : msg.get("From"),
        'To' : msg.get("To"),
        'Cc' : msg.get("Cc"),
        'Bcc' : msg.get("Bcc"),
        'Subject': subject,
        'Body': body,
        'Attachements': attachments,
        'Alias' : self.alias,
        }
        # ----------------------------------------------------

        if delete: self.imap.store(message_id, '+FLAGS', '\\Deleted')
        if flag: self.imap.store(message_id, '+FLAGS', '\\Flagged')
        if seen: self.imap.store(message_id, '+FLAGS', '\\Seen')

        return res
    
    def CloseFolder(self):
        if self.imap.state=='SELECTED':
            self.imap.expunge()
            self.imap.close()
            self.folder=None
            self.messages=None
            return True
        else: return False

    def Logout(self): 
        self.imap.logout() # will call self.imap.shutdown()
        self.Reset()
    
    # -------------------------------------------------------------------------------------

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=




__doc__=r"""pix module

Basic Image manupulation from the command line, 

```
python -m knwon.pix --help

options:
  -h, --help         show this help message and exit
  --action ACTION    (str) one of the static-methods inside the Actions class, 
                           can be - ['new', 'crop', 'extend', 'flip', 'rotate', 'convert', 'autoconvert']
  --args ARGS        (str) csv args accepted by the specified action - each action takes different args
  --input INPUT      (str) input image-file or a text/json-file containing multiple input image-file names
  --output OUTPUT    (str) output image-file or a text/json-file containing multiple output image-file names
  --files FILES      (str) multiple input image-file names - for custom action -- works only with --io=linux
  --io IO            (str) can be 'text' or 'json' or 'linux' - keep blank to io as 'image' - used if providing 
                           input/output file-names in a text/json file
  --verbose VERBOSE  (int) verbose level - 0 or 1

```

## Functionality

#### 1. Create a new image

* creates new images of given size and color 

* args is (int) 7-tuple (height, width, channels, blue, green, red, alpha)
    * `--args=h,w,c,b,g,r,a`

* fill-color (blue, green, red, alpha) depends on the specified number of channels
    * `--args=h,w,4,b,g,r,a`
    * `--args=h,w,3,b,g,r`
    * `--args=h,w,1,i`                  

Example - creating 32 x 64 image with 4 channels named 'new.jpg'
```
python -m known.pix --action=new --args=32,64,4,100,150,200,170 --output=new.jpg
```

#### 2. Crop an image

* crops an image using bounding box (y, x, h, w) 

* args is (int) 4-tuple (y-coord, x-coord, height, width) indicating a bounding box
    * `--args=y,x,h,w`

Example - 
```
python -m known.pix --action=crop --args=8,16,16,32 --input=new.jpg --output=cropped.jpg
```

#### 3. Extend an image

* extends an image using boundary distance 

* args is (int) 8-tuple (north, south, east, west, blue, green, red, alpha)
    * `--args=n,s,e,w,b,g,r,a`

* fill-color (blue, green, red, alpha) depends on the specified number of channels
    * `--args=n,s,e,w,b,g,r,a`
    * `--args=n,s,e,w,b,g,r`    
    * `--args=n,s,e,w,i`

Example - 
```
python -m known.pix --action=extend --args=10,5,6,12,123,123,0,100 --input=new.jpg --output=extended.jpg
```


#### 4. Flip an image

* flip an image (horizontally, vertically)

* args is (int) 2-tuple (horizontally, vertically) 

    * Flip horizontally          `--args=1,0`
    * Flip vertically            `--args=0,1`
    * Flip corners               `--args=1,1`
    * Flip nothing               `--args=0,0`

Example - 
```
python -m known.pix --action=flip --args=1,1 --input=extended.jpg --output=flipped.jpg
```

#### 5. Rotate an image

* rotate an image (clockwise or couter-clockwise)

* args is (int) 1-tuple (clockwise) 
    
    * Rotate clockwise               `--args=1`
    * Rotate counter-clockwise       `--args=0`

Example - 
```
python -m known.pix --action=rotate --args=1 --input=flipped.jpg --output=rotated.jpg
```

#### 6. Convert an image to a format

* converts between image formats *(as per output)*

* args is not used, target file type is infered from the output file extension

Example - 
```
python -m known.pix --action=convert --input=rotated.jpg --output=rotated.png
```

#### 7. Convert an image to multiple formats

* converts between image formats *(as per args)*

* args is (str) n-tuple specifying the extensions to be converted to

* output filenames are not used, the file-names are taken from input files

* the extensions are added as specified in args

* e.g., Convert png to jpg and webp      `--input=input.png --args=jpg,webp`
the output files will be `input.png` and `input.webp`

Example - 
```
python -m known.pix --action=autoconvert --input=rotated.png --args=jpeg,webp
```
"""

import os
import numpy as np
import cv2

DTYPE =             np.uint8
VALID_CHANNELS =    set([1, 3, 4])
DEFAULT_FILL =      255


# returns a new array
def create(h:int, w:int, c:int):
    # creates a new ndarray of specified height, width and channels with datatype specified in class (no of channels can be 1, 3 or 4)
    assert c in VALID_CHANNELS, f'[{c}] is not a valid number of channels, expecting {VALID_CHANNELS}'
    return np.zeros((h, w, c), dtype=DTYPE) 

def save(pix, path):  return cv2.imwrite(path, pix)

# returns a new array
def load(path): 
    img =  cv2.imread(path, cv2.IMREAD_UNCHANGED) #<--- can throw error
    if img.ndim==3:     pix = img.astype(DTYPE)
    elif img.ndim==2:   pix = np.expand_dims(img.astype(DTYPE), -1)
    else:               raise ValueError(f'expecting 2/3-D array but got {img.ndim}-D')
    return pix

# returns a view of the array
def flip(pix, horizontal=False, vertical=False): return pix[(slice(None, None, -1) if vertical else slice(None, None, None)), (slice(None, None, -1) if horizontal else slice(None, None, None)), :]

# returns a view of the array
def rotate(pix, clockwise=True):  return np.swapaxes(pix, 0, 1)[:, ::-1, :] if clockwise else np.swapaxes(pix, 0, 1)[::-1,:,:]

# in-place method
def fill(pix, i_row:int, i_col:int, n_rows:int, n_cols:int, color:tuple, channel=None): 
    # fills an area (region) of the image on all channels or a particular channel with values provided in color
    # i_row, i_col are the starting row and col
    # n_row, n_col are the number of rows and cols
    # if channel is provided, then color is assumed to be integer
    # otherwise color should be a tuple of the form (bgra in 4 channel), (bgr in 3 channel), (intensity in 1 channel)
    if channel is None: 
        ic = slice(None, None, None) 
        if not color: color = [ DEFAULT_FILL for _ in range(pix.shape[-1]) ]   
        else: color = color[:pix.shape[-1]]
    else:
        ic = channel # abs(int(channel)) % pix.shape[-1]
        if color is None: color = DEFAULT_FILL 

    if (n_rows is ...) or (n_rows is None): 
        if (n_cols is ...) or (n_cols is None):     pix[i_row:,             i_col:,             ic] =  color
        else:                                       pix[i_row:,             i_col:i_col+n_cols, ic] =  color
    else: 
        if (n_cols is ...) or (n_cols is None):     pix[i_row:i_row+n_rows, i_col:,             ic] =  color
        else:                                       pix[i_row:i_row+n_rows, i_col:i_col+n_cols, ic] =  color

# in-place method
def region_fill(pix, start_row:int, start_col:int, n_rows:int, n_cols:int, color:tuple):  fill(pix, start_row, start_col, n_rows, n_cols, color, channel=None)

# in-place method
def region_copy(pix_from, start_row_from, start_col_from, n_rows, n_cols, pix_to, start_row_to, start_col_to): pix_to[start_row_to:start_row_to+n_rows, start_col_to:start_col_to+n_cols,  :] = pix_from[start_row_from:start_row_from+n_rows, start_col_from:start_col_from+n_cols, :]

# returns a new array
def extend(pix, north, south, east, west, filler=None):
    # extends an image on all four sides (specified by number of pixels in north, south, east, west)
    new = create(pix.shape[0] + (abs(north) + abs(south)), pix.shape[1] + (abs(east) + abs(west)), pix.shape[-1])
    region_fill(new, 0, 0, None, None, filler)
    region_copy(
        pix_from = pix,   
        start_row_from = 0,
        start_col_from = 0,           
        n_rows = pix.shape[0],
        n_cols = pix.shape[1],
        pix_to = new,     
        start_row_to = abs(north),
        start_col_to = abs(west))
    return new


class Actions:
    r"""

    Defins the available actions to this module (refer __main__.py)

        python -m known.pix --action=<member-function>

    Note: 
        (blue, green, red, alpha) can take values from 0 to 255
        (channels) can be 4, 3 or 1 - the color tuple depends on the channels in the image
    """
    @staticmethod
    def new(inputs, outputs, args, verbose=0, dry=False):
        
        """ creates new images of given size and color 
        
        args is (int) 7-tuple (height, width, channels, blue, green, red, alpha)

            --args=h,w,c,b,g,r,a
        
        fill-color (blue, green, red, alpha) depends on the specified number of channels

            --args=h,w,4,b,g,r,a
            --args=h,w,3,b,g,r            
            --args=h,w,1,i                    
        """
        if verbose: print(f'⚙ [NEW] {len(outputs)}')
        try:
            args = [int(s) for s in args]
            if verbose: print(f'● Dimension {args[0:3]} with fill-color {args[3:]}')
            ip='new'
            for op in outputs:
                if verbose: print(f'\t● {ip} ⇒ {op}')
                img = create(*args[0:3])
                fill(img, 0, 0, ..., ..., color=args[3:], channel=None)
                if dry: print(f'\n[DRY ~ SAVE]\n\t\n\tIMAGE: {img.shape}\n\tOUTPUT: {op}\n')
                else: save(img, op)
            if verbose: print(f'✓ Success!')
        except:
            if verbose: print(f'✗ Failed!')
        
    @staticmethod
    def crop(inputs, outputs, args, verbose=0, dry=False):
        """ crops an image using bounding box (y, x, h, w) 

        args is (int) 4-tuple (y-cord, x-coord, height, width) indicating a bounding box
            
            --args=y,x,h,w
        """
        if verbose: print(f'⚙ [CROP] {len(inputs)}')
        try:
            y, x, h, w = [int(s) for s in args]
            if verbose: print(f'● Bounding-Box {y=} {x=} {h=} {w=}')
            for ip,op in zip(inputs,outputs):
                if verbose: print(f'\t● {ip} ⇒ {op}')
                org = load(ip)
                img = create(h, w, org.shape[-1])
                region_copy(org, y, x, h, w, img, 0, 0)
                if dry: print(f'\n[DRY ~ SAVE]\n\tINPUT: {ip}\n\tIMAGE: {org.shape} --> {img.shape}\n\tOUTPUT: {op}\n')
                else: save(img, op)
            if verbose: print(f'✓ Success!')
        except:
            if verbose: print(f'✗ Failed!')

    @staticmethod
    def extend(inputs, outputs, args, verbose=0, dry=False):
        """ extends an image using boundary distance 

        args is (int) 8-tuple (north, south, east, west, blue, green, red, alpha)
            
            --args=n,s,e,w,b,g,r,a

        fill-color (blue, green, red, alpha) depends on the specified number of channels

            --args=n,s,e,w,b,g,r,a
            --args=n,s,e,w,b,g,r    
            --args=n,s,e,w,i
        """
        if verbose: print(f'⚙ [EXTEND] {len(inputs)}')
        try:
            args = [int(s) for s in args]
            north, south, east, west = args[0:4]
            if verbose: print(f'● Directions {north=} {south=} {east=} {west=}')
            for ip,op in zip(inputs,outputs):
                if verbose: print(f'\t● {ip} ⇒ {op}')
                img = extend(load(ip), north, south, east, west, filler=args[4:])
                if dry: print(f'\n[DRY ~ SAVE]\n\tINPUT: {ip}\n\tIMAGE: {img.shape}\n\tOUTPUT: {op}\n')
                else: save(img, op)
            if verbose: print(f'✓ Success!')
        except:
            if verbose: print(f'✗ Failed!')

    @staticmethod
    def flip(inputs, outputs, args, verbose=0, dry=False):
        """ flip an image (horizontally, vertically)

        args is (int) 2-tuple (horizontally, vertically) 
        
            Flip horizontally          --args=1,0
            Flip vertically            --args=0,1
            Flip corners               --args=1,1
            Flip nothing               --args=0,0
        """
        if verbose: print(f'⚙ [FLIP] {len(inputs)}')
        try:
            h, v = [bool(int(s)) for s in args]
            if verbose: print(f'● Directions {h=} {v=}')
            for ip,op in zip(inputs,outputs):
                if verbose: print(f'\t● {ip} ⇒ {op}')
                org = load(ip)
                img = flip(org, horizontal=h, vertical=v)
                if dry: print(f'\n[DRY ~ SAVE]\n\tINPUT: {ip}\n\tIMAGE: {org.shape} --> {img.shape}\n\tOUTPUT: {op}\n')
                else: save(img, op)
            if verbose: print(f'✓ Success!')
        except:
            if verbose: print(f'✗ Failed!')

    @staticmethod
    def rotate(inputs, outputs, args, verbose=0, dry=False):
        """ rotate an image (clockwise or couter-clockwise)

        args is (int) 1-tuple (clockwise) 
            
            Rotate clockwise               --args=1
            Rotate counter-clockwise       --args=0
        """
        if verbose: print(f'⚙ [ROTATE] {len(inputs)}')
        try:
            c = [bool(int(s)) for s in args][0]
            if verbose: print(f'● Direction clockwise - {c}')
            for ip,op in zip(inputs,outputs):
                if verbose: print(f'\t● {ip} ⇒ {op}')
                org = load(ip)
                img = rotate(org, clockwise=c)
                if dry: print(f'\n[DRY ~ SAVE]\n\tINPUT: {ip}\n\tIMAGE: {org.shape} --> {img.shape}\n\tOUTPUT: {op}\n')
                else: save(img, op)
            if verbose: print(f'✓ Success!')
        except:
            if verbose: print(f'✗ Failed!')
    
    @staticmethod
    def convert(inputs, outputs, args, verbose=0, dry=False):
        """ converts an image (as per output)
        
        args is not used, target file type is infered from the output file-name (extension)
            
        e.g., Convert png to jpg      --input=input.png --output=output.jpg
        """
        
        if verbose: print(f'⚙ [CONVERT] {len(inputs)}')
        try:
            for ip,op in zip(inputs,outputs): 
                if verbose: print(f'\t● {ip} ⇒ {op}')
                img = load(ip)
                if dry: print(f'\n[DRY ~ SAVE]\n\tINPUT: {ip}\n\tIMAGE: {img.shape}\n\tOUTPUT: {op}\n')
                else: save(img, op)
            if verbose: print(f'✓ Success!')
        except:
            if verbose: print(f'✗ Failed!')

    @staticmethod
    def autoconvert(inputs, outputs, args, verbose=0, dry=False):
        """ converts an image (as per args)
        
        args is (str) n-tuple specifying the extensions to be converted to
        output filenames are not used, the file-names are taken from input files
        the extensions are added as specified in args

        e.g., Convert png to jpg and webp      --input=input.png --args=jpg,webp
        the output files will be `input.jpg` and `input.webp`

        """
        if verbose: print(f'⚙ [AUTOCONVERT] {len(inputs)}')
        try:
            for ip in inputs: 
                dirname, filename = os.path.dirname(ip), os.path.basename(ip)
                i = filename.rfind('.')
                name = filename if i<0 else filename[0:i]
                for ext in args:
                    op = os.path.join(dirname, f'{name}.{ext}')
                    if verbose: print(f'\t● {ip} ⇒ {op}')
                    img = load(ip)
                    if dry: print(f'\n[DRY ~ SAVE]\n\tINPUT: {ip}\n\tIMAGE: {img.shape}\n\tOUTPUT: {op}\n')
                    else: save(img, op)
            if verbose: print(f'✓ Success!')
            return True
        except:
            if verbose: print(f'✗ Failed!')
            return False

    @staticmethod
    def multiconvert(inputs, outputs, args, verbose=0, dry=False):
        """ converts all the images in a directory (as per args)
        
        input is a directory name
        args is csv source extension and target extension ending with .

        e.g., Convert png and bmp to jpg and webp      --input=/home/user --args=.,png,bmp,jpg.,webp.

        """
        args = set(args)
        remove_after_convert = False
        if "." in args: 
            remove_after_convert=True
            args.remove(".")
        tgt_ext = tuple(set([e[:-1] for e in args if e.endswith('.')]))
        src_ext = tuple(set([e      for e in args if not e.endswith('.')]))
        
        if verbose: print(f'⚙ [MULTICONVERT] {inputs}')
        for in_dir in inputs:
            input_files = [os.path.join(in_dir, f) for f in os.listdir(in_dir) if f'{f}'.lower().endswith(src_ext)]
            res = __class__.autoconvert(
                inputs= input_files,
                outputs=None,
                args=tgt_ext,
                verbose=verbose,
                dry=dry,
            )
            if res:
                if verbose: print(f'✓ [MULTICONVERT] Success for {in_dir}')
                if remove_after_convert: 
                    if verbose: print(f'\t ~ Deleting {len(input_files)} Files...')
                    if dry:
                        for f in input_files: print(f'Delete file {f}')
                    else: 
                        for f in input_files: os.remove(f)
            else:
                if verbose: print(f'✗ [MULTICONVERT] Failed for {in_dir}')



__doc__="""
Generic functions for pytorch modules
"""

import torch as tt
import torch.nn as nn
from io import BytesIO


def numel(shape): 
    r""" returns no of total elements (or addresses) in a multi-dim array 
        Note: for torch tensor use Tensor.numel()"""
    return tt.prod(tt.tensor(shape)).item()

def arange(shape, start=0, step=1, dtype=None): 
    r""" returns arange for multi-dimensional array (reshapes) """
    return tt.arange(start=start, end=start+step*numel(shape), step=step, dtype=dtype).reshape(shape)

def shares_memory(a, b) -> bool: 
    r""" checks if two tensors share same underlying storage, in which case, changing values of one will change values in other as well
        Note: this is different from Tensor.is_set_to(Tensor) function which checks shape as well"""
    return (a.storage().data_ptr() == b.storage().data_ptr())

def absdiff(a, b): return tt.sum(tt.abs(a-b)).item()

def save_state(path, module): tt.save(module.state_dict(), path) # simply save the state dictionary

def load_state(path, module): module.load_state_dict(tt.load(path)) # simply load the state dictionary

def save_to_buffer(module, seek0=False): 
    buffer = BytesIO()
    tt.save(module, buffer)
    if seek0: buffer.seek(0)
    return buffer

def load_from_buffer(buffer, seek0=True): 
    if seek0: buffer.seek(0)
    return tt.load(buffer)

def save(module, path:str): tt.save(module, path)

def load(path:str): return tt.load(path)

def count(module, requires_grad=None): 
    r""" Counts the total number of parameters (numel) in a params
    
    :param requires_grad: 
        if None, counts all parameters
        if True, counts trainable parameters
        if False, counts non-trainiable (frozen) parameters
    """
    return sum( ([ p.numel() for p in module.parameters() ]) if requires_grad is None else \
                ([ p.numel() for p in module.parameters()    if p.requires_grad is requires_grad ]) )

def show(module, values:bool=False):
    r""" Prints the parameters of a params
    
    :param values: if True, prints the full tensors otherwise prints only shape
    """
    nos_trainable, nos_frozen = 0, 0
    print('=====================================')
    for i,p in enumerate(module.parameters()):
        iparam = p.numel()
        if p.requires_grad:
            nos_trainable += iparam
        else:
            nos_frozen += iparam
        print(f'#[{i}]\tShape[{p.shape}]\tParams: {iparam}\tTrainable: {p.requires_grad}')
        if values: 
            print('=====================================')
            print(f'{p}')
            print('=====================================')
    print(f'\nTotal Parameters: {nos_trainable+nos_frozen}\tTrainable: {nos_trainable}\tFrozen: {nos_frozen}')
    print('=====================================')
    return 

def state(module, values=False):
    r""" prints the parameters using `nn.Module.parameters` iterator, use `values=True` to print full parameter tensor """
    sd = module.state_dict()
    for i,(k,v) in enumerate(sd.items()):
        print(f'#[{i+1}]\t[{k}]\tShape[{v.shape}]')
        if values: print(f'{v}')
    return 

@tt.no_grad()
def diff(module1, module2, do_abs:bool=True, do_sum:bool=True):
    r""" Checks the difference between the parameters of two modules.
        This can be used to check if two models have exactly the same parameters.

    :param do_abs: if True, finds the absolute difference
    :param do_sum: if True, finds the sum of difference

    :returns: a list of differences in each parameter or their sum if ``do_sum`` is True.
    """
    d = [ (abs(p1 - p2) if do_abs else (p1 - p2)) for p1,p2 in zip(module1.parameters(), module2.parameters()) ]
    if do_sum: d = [ tt.sum(p) for p in d  ]
    return d

@tt.no_grad()
def copy(module_from, module_to) -> None:
    r""" Copies the parameters of a params to another - both modules are supposed to be identical"""
    for pt,pf in zip(module_to.parameters(), module_from.parameters()): pt.copy_(pf)

def clones(module, n_copies:int):
    r""" Replicates a params by storing it in a buffer and retriving many copies
    NOTE: this will preserve the ```require_grad``` attribute on all tensors. """
    #from io import BytesIO
    if n_copies<1: return None
    buffer = BytesIO()
    tt.save(module, buffer)
    model_copies = []
    for _ in range(n_copies):
        buffer.seek(0)
        model_copy = tt.load(buffer)
        model_copies.append(model_copy)
    buffer.close()
    del buffer
    return model_copies

def clone(module): return clones(module, 1).pop()

def duplicate(module, n_copies): return nn.ModuleList(clones(module, n_copies))

def requires_grad_(module, requires:bool, *names):
    r""" Sets requires_grad attribute on tensors in params
    if no names are provided, sets requires_grad on all tensors 
    NOTE: careful with *names, if a buffer's name is provided
        and it is in the state_dict then its grad will be enabled
        which is undesirable.
        not providing any names will target the parameters only
    """
    if names: # if we know which params to freeze, we can provide them
        state_dict = module.state_dict() 
        for n in names: state_dict[n].requires_grad_(requires)
    else: # if we want to do on all params
        for p in module.parameters(): p.requires_grad_(requires)
    return module

@tt.no_grad()
def zero_(module, *names):
    r""" Sets requires_grad attribute on tensors in params
    if no names are provided, sets requires_grad on all tensors 
    
    NOTE: careful with *names, if a buffer's name is provided
        and it is in the state_dict then it will be zeroed too
        which is actually desirable in some cases.
        pass a single blank string to zero everything in state dict
        not providing any names will target the parameters only
    """
    if names:
        state_dict = module.state_dict()
        if " " in names:
            for p in state_dict.values(): p.zero_() 
        else:   
            for n in names: state_dict[n].zero_()
    else: 
        for p in module.parameters(): p.zero_()
    return module

def zero_like(module) -> dict: return zero_(clone(module), " ")

def dense(in_dim, layer_dims, out_dim, 
        actFs, bias=True, dtype=None, device=None ):
    r"""
    Creats a stack of fully connected (dense) layers which is usually connected at end of other networks
    Args:
        in_dim          `integer`       : in_features or input_size
        layer_dims      `List/Tuple`    : size of hidden layers
        out_dim         `integer`       : out_features or output_size
        actFs           `nn.Module`     : activation function at hidden layer
        bias            `bool`          : if True, uses bias at hidden layers

    Returns:
        `nn.Module` : an instance of nn.Sequential
    """
    layers = []
    # first layer
    layers.append(nn.Linear(in_dim, layer_dims[0], bias=bias, dtype=dtype, device=device))
    if actFs: layers.append(actFs.pop(0))
    # remaining layers
    for i in range(len(layer_dims)-1):
        layers.append(nn.Linear(layer_dims[i], layer_dims[i+1], bias=bias, dtype=dtype, device=device))
        if actFs: layers.append(actFs.pop(0))
    # last layer
    layers.append(nn.Linear(layer_dims[-1], out_dim, bias=bias, dtype=dtype, device=device))
    if actFs: layers.append(actFs.pop(0))
    return nn.Sequential( *layers )

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=



__doc__=""" Helper Functions and Utils """

import os, requests

__all__ =['PublicIPv4', 'ParseLinuxFiles', 'ConfigParser', 'ImportCustomModule', 'GraphFromImage', 'Int2File', 'File2Int',
        'NewNotebook', 'NB2HTML']


def PublicIPv4():
    r""" gets your public IPv4 from api.ipify.org"""
    try:public_ip = requests.get('https://api.ipify.org').text
    except requests.exceptions.RequestException as e: public_ip = None
    return public_ip

def ParseLinuxFiles(F, check=False): # parses --files="%F"
    Fl = [fi.strip() for fi in F.split("'/")]
    Fr = [os.path.abspath(f'/{fl[:-1]}'.replace("'\\''","'")) for fl in Fl if fl] 
    if check: Fr = [fr for fr in Fr if os.path.exists(fr)]
    return Fr

def ConfigParser(dict_to_object=True):
    r""" A replacement for usual argparse method of passing arguments to python scripts

    The problems are:
        1. A large number of arguments makes it messy to run everytime
        2. Arguments can be passed as premitive types only - we want to use any python object like ndarray, tensors and dataframes too

    The soultion is - use config files instead of arguments
        > A config file is a python scripts that contains config objects
        > This config file can be passed to a script like `python script.py --config=/path/to/config`
        > There may be multiple configs inside a single config file, hence the specific config to choose must be provided by `--member`
        > Since its a python script that will be imported on call so that we can create objects dynamically


    Example usage:

    [1] In the script file (`/path/to/script.py`) that has to be run, enter the following code

    ```python

        from known.utils import ConfigParser
        args = ConfigParser() # these are the args

    ```

    [2] In the config file ('/path/to/config.py'), enter the following code

    ```python

        import numpy as np
        import os, datetime

        class MyArgs:
            # -------------------------------------------------------------------------------------------------
            # when defining args in a class like this, it will executed as soon as this file is imported
            # -------------------------------------------------------------------------------------------------
            path =      f'{__file__}'
            cwd =       os.getcwd()
            name =      "Monty Python"
            intarg =    19
            floatarg =  99.99
            listarg =   [1, 2.0, "3"]
            tuplearg =  (listarg, intarg, floatarg)
            dictarg =   dict(place="Earth", position=(1.09, 2.908))
            nparg =     np.arange(intarg)
            timearg =   datetime.datetime.now()
            # -------------------------------------------------------------------------------------------------

        def MyArgc():
            # -------------------------------------------------------------------------------------------------
            # when defining args in a function like this, it should not accept any arguments
            # -------------------------------------------------------------------------------------------------
            return dict (
                path =      f'{__file__}',
                cwd =       os.getcwd(),
                name =      "Monty Python",
                intarg =    19,
                floatarg =  99.99,
                listarg =   [1, 2.0, "3"],
                dictarg =   dict(place="Earth", position=(1.09, 2.908)),
                nparg =     np.arange(45),
                timearg =   datetime.datetime.now(),
            )
            # -------------------------------------------------------------------------------------------------

    ```

    [3] Call the script file with following arguments:

        [3.1] either use a class that itself is a config object (class `MyArgs` in this example)

            python /path/to/script.py --config=/path/to/config.py:MyArgs

        # note the semi-colon (:) to indicate that MyArgs is the user config object

        [3.2] or use a callable function that returns a config object or a dict (function `MyArgc` in this example)

            python /path/to/script.py --config=/path/to/config.py:MyArgc!

        # note the exclamation mark (!) to indicate that MyArgc should be called (with no args)

    [NOTE]

        # the `dict_to_object` argument is only valid when user returns a dict (instead of class or object or callable)
        # if user returns anything other than a dict, it will be passed as it is - value of `dict_to_object` will not matter
        
        # if `dict_to_object==True`,    converts a dict (returned by user) to an object (this object is an instance of class 'FakeConfig') 
        # otherwise,                    it will pass the dict as it is, which can be accessed by keys only

        # this provides flexibility to users as they can define arguments and access them in both possible ways
        # a noteable case is when the arguments names are not valid variable names in python  
        # access the dict using `FakeConfig.__dict__` from a `FakeConfig` object
        
    """

    import os, sys, importlib, argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--config', type=str, default='')
    parsed = parser.parse_args()
    

    config_string = f'{parsed.config}'
    if not config_string: raise RuntimeError(f'Configuration is missing!')
    # first check if it ends in '!'
    provided_called = config_string.endswith("!")
    if provided_called: config_string = config_string[:-1]
    if not config_string: raise RuntimeError(f'Configuration is incomplete! {config_string}')
    # now find the callable by reading backward untill ':' is encountered
    split_at = config_string.rfind(":")
    if split_at<0: raise RuntimeError(f'Configuration is incomplete - member (:) not provided! {config_string}')
    CONFIG_MEMBER = config_string[split_at+1:]
    provided_config = config_string[:split_at]

    if not CONFIG_MEMBER:                               raise NameError(f'Member name is missing!')
    if not os.path.isfile(provided_config):             raise FileNotFoundError(f'Config Module File not found at [{provided_config}]')
    if not provided_config.lower().endswith('.py'):     raise RuntimeError(f'Expecting python script (.py) but got {provided_config}')
    #-----------------------------------------------------------------------------------------
    # ==> read configurations
    #-----------------------------------------------------------------------------------------
    CONFIG_FILE_PATH = os.path.abspath(provided_config)
    CONFIG_DIR_PATH, CONFIG_MODULE_NAME = os.path.dirname(CONFIG_FILE_PATH), os.path.basename(CONFIG_FILE_PATH)[:-3]
    if CONFIG_DIR_PATH not in sys.path: sys.path.append(CONFIG_DIR_PATH) 
    #<---- appends directory of config file to sys.path ->> this will add it to sys.path untill logout
    try:    CONFIG_MODULE = importlib.import_module(CONFIG_MODULE_NAME)
    except: raise ModuleNotFoundError(f'Could import module "{CONFIG_MODULE_NAME}" at file "{CONFIG_FILE_PATH}"')
    try:    CONFIG_OBJECT = getattr(CONFIG_MODULE, CONFIG_MEMBER)
    except: raise RuntimeError(f'Could not find {CONFIG_MODULE_NAME}.{CONFIG_MEMBER}')
    if provided_called:
        try:    CONFIG_OBJECT = CONFIG_OBJECT()
        except: raise RuntimeError(f'Error calling {CONFIG_MODULE_NAME}.{CONFIG_MEMBER} - is the config correct?')

    if dict_to_object and isinstance(CONFIG_OBJECT, dict):
        class FakeConfig:
            def __len__(self): return len(self.__dict__)
            def __contains__(self, x): return x in self.__dict__
            def _get_kwargs(self): return self.__dict__.items() # mimic inbuilt ArgumentParser
            def _get_args(self): return self.__dict__.values()  # mimic inbuilt ArgumentParser
            def __init__(self, **kwargs):
                for k,v in kwargs.items(): setattr(self, k, v)

        try:    CONFIG_OBJECT = FakeConfig(**CONFIG_OBJECT)
        except: raise RuntimeError(f'Could not create python-object from dict {CONFIG_MODULE_NAME}.{CONFIG_MEMBER}.{CONFIG_OBJECT} - are the arguments named properly?')

    return CONFIG_OBJECT #<---- this can return a dict as well (set the argument dict_to_object = False )

def ImportCustomModule(python_file:str, python_object:str='', do_initialize:bool=False):
    r""" Import a custom module from a python file and optionally initialize it """
    import os, importlib.util
    cpath = os.path.abspath(python_file)
    failed=""
    if os.path.isfile(cpath): 
        try: 
            # from https://stackoverflow.com/questions/67631/how-can-i-import-a-module-dynamically-given-the-full-path
            cspec = importlib.util.spec_from_file_location("", cpath)
            cmodule = importlib.util.module_from_spec(cspec)
            cspec.loader.exec_module(cmodule)
            success=True
        except: success=False #exit(f'[!] Could import user-module "{cpath}"')
        if success: 
            if python_object:
                try:
                    cmodule = getattr(cmodule, python_object)
                    if do_initialize:  cmodule = cmodule()
                except:         cmodule, failed = None, f'[!] Could not import object {python_object} from module "{cpath}"'
        else:                   cmodule, failed = None, f'[!] Could not import module "{cpath}"'
    else:                       cmodule, failed = None, f"[!] File Not found @ {cpath}"
    return cmodule, failed

def GraphFromImage(img_path:str, pixel_choice:str='first', dtype=None):
    r""" 
    Covert an image to an array (1-Dimensional)

    :param img_path:        path of input image 
    :param pixel_choice:    choose from ``[ 'first', 'last', 'mid', 'mean' ]``

    :returns: 1-D numpy array containing the data points

    .. note:: 
        * This is used to generate synthetic data in 1-Dimension. 
            The width of the image is the number of points (x-axis),
            while the height of the image is the range of data points, choosen based on their index along y-axis.
    
        * The provided image is opened in grayscale mode.
            All the *black pixels* are considered as data points.
            If there are multiple black points in a column then ``pixel_choice`` argument specifies which pixel to choose.

        * Requires ``opencv-python``

            Input image should be readable using ``cv2.imread``.
            Use ``pip install opencv-python`` to install ``cv2`` package
    """
    import cv2
    import numpy as np
    img= cv2.imread(img_path, 0)
    imgmax = img.shape[1]-1
    j = img*0
    j[np.where(img==0)]=1
    pixel_choice = pixel_choice.lower()
    pixel_choice_dict = {
        'first':    (lambda ai: ai[0]),
        'last':     (lambda ai: ai[-1]),
        'mid':      (lambda ai: ai[int(len(ai)/2)]),
        'mean':     (lambda ai: np.mean(ai))
    }
    px = pixel_choice_dict[pixel_choice]
    if dtype is None: dtype=np.float_
    return np.array([ imgmax-px(np.where(j[:,i]==1)[0]) for i in range(j.shape[1]) ], dtype=dtype)

def File2Int(file, col=None):
    r""" Reads a file and converts bytes to a list of integers - this is useful for harcoding small files"""
    with open(file, 'rb') as f: B = f.read()
    I = [int(b) for b in B]
    if col:
        print('[')
        for c,i in enumerate(I,1): 
            print(f'{i},', end="")
            if not c%col: print('') 
        print('\n]')
    return I

def Int2File(I, file):
    r""" reads bytes from list of Intergers (I) and writes to file """
    with open(file, 'wb') as f:
        for i in I: f.write(i.to_bytes())


def NewNotebook(heading="# Notebook", nbformat=4, nbformat_minor=2, save=""): 
    r""" Create a new empty notebook and optionally saves it"""
    res = f"""
{{
    "cells": 
        [
            {{
                "cell_type": "markdown",
                "metadata": {{}},
                "source": [ "{heading}" ] 
            }} 
        ],
    "metadata": {{}}, 
    "nbformat": {nbformat}, 
    "nbformat_minor": {nbformat_minor}
}}
"""
    if save: 
        with open(save, 'w') as f: f.write(res)
    return res

def NB2HTML(source_notebook, template_name='lab', no_script=False):
    r""" Converts a notebook to html with added name (title) and template, optionally removes any scripts"""
    from bs4 import BeautifulSoup
    from nbconvert import HTMLExporter
    page, _ = HTMLExporter(template_name=template_name) \
            .from_file(source_notebook, dict(metadata=dict(name = f'{os.path.basename(source_notebook)}')),) 
    soup = BeautifulSoup(page, 'html.parser')
    if no_script: # Find all script tags and remove them
        for script in soup.find_all('script'): script.decompose()  
    return soup.prettify()




from setuptools import setup, find_packages

setup(
    name =                      'known',
    version =                   '0.0.32',
    url =                       'https://github.com/auto-notify-ps/known',
    author =                    'Nelson.S',
    author_email =              'mail.nelsonsharma@gmail.com',
    description =               'a collection of reusable python code',
    packages =                  find_packages(include=['known', 'known.*']),
    classifiers=                ['License :: OSI Approved :: MIT License'],
    #package_dir =               { '' : ''},
    install_requires =          [],
    include_package_data =      False,
    #python_requires =           ">=3.8",
)   